#! /usr/bin/env python
# -*- coding: utf-8 -*-

import torch
import pickle

import os.path as osp
import numpy as np
import pandas as pd
import scipy.sparse as sp

from torch_geometric.data import Data
from torch_sparse import coalesce
from sklearn.feature_extraction.text import CountVectorizer

def load_dataset(path='../hyperGCN/data/', dataset = 'cora', train_percent = 0.025):
    
    # first load node features:
    with open(osp.join(path, dataset, 'features.pickle'), 'rb') as f:
        features = pickle.load(f)
        if dataset in ['actor','amazon','pokec','twitch']:
            pass
        else:
            features = features.todense()

    # then load node labels:
    with open(osp.join(path, dataset, 'labels.pickle'), 'rb') as f:
        labels = pickle.load(f)

    num_nodes, feature_dim = features.shape
    assert num_nodes == len(labels)
    print(f'number of nodes:{num_nodes}, feature dimension: {feature_dim}')

    features = torch.FloatTensor(features)
    labels = torch.LongTensor(labels)

    # The last, load hypergraph.
    with open(osp.join(path, dataset, 'hypergraph.pickle'), 'rb') as f:
        # hypergraph in hyperGCN is in the form of a dictionary.
        # { hyperedge: [list of nodes in the he], ...}
        hypergraph = pickle.load(f)

    print(f'number of hyperedges: {len(hypergraph)}')

    edge_idx = num_nodes
    node_list = []
    edge_list = []
    for he in hypergraph.keys():
        cur_he = hypergraph[he]
        cur_size = len(cur_he)

        node_list += list(cur_he)
        edge_list += [edge_idx] * cur_size

        edge_idx += 1

    edge_index = np.array([ node_list + edge_list,
                            edge_list + node_list], dtype = int)
    edge_index = torch.LongTensor(edge_index)

    data = Data(x = features,
                edge_index = edge_index,
                y = labels)

    total_num_node_id_he_id = edge_index.max() + 1
    data.edge_index, data.edge_attr = coalesce(data.edge_index, 
            None, 
            total_num_node_id_he_id, 
            total_num_node_id_he_id)
            
    n_x = num_nodes
    data.n_x = n_x

    data.train_percent = train_percent
    data.num_hyperedges = len(hypergraph)
    
    return data

def load_large_dataset(path='../hyperGCN/data/', dataset = 'trivago', train_percent = 0.025):
    
    features = torch.load(osp.join(path,dataset,f'{dataset}_X_vec.pt'))

    # then load node labels:
    labels = torch.load(osp.join(path,dataset,f'{dataset}_Y.pt'))

    num_nodes, feature_dim = features.shape
    assert num_nodes == len(labels)
    print(f'number of nodes:{num_nodes}, feature dimension: {feature_dim}')

    hypergraph = torch.load(osp.join(path,dataset,f'{dataset}_E.pt'))

    num_hyperedges = hypergraph[1].max()+1
    print(f'number of hyperedges: {hypergraph[1].max()+1}')

    edge_idx = num_nodes
    hypergraph[1]+=edge_idx 
    node_list,edge_list = hypergraph[0].tolist(),hypergraph[1].tolist()

    edge_index = np.array([ node_list + edge_list,
                            edge_list + node_list], dtype = int)
    
    edge_index = torch.LongTensor(edge_index)

    data = Data(x = features,
                edge_index = edge_index,
                y = labels)

    total_num_node_id_he_id = edge_index.max() + 1
    data.edge_index, data.edge_attr = coalesce(data.edge_index, 
            None, 
            total_num_node_id_he_id, 
            total_num_node_id_he_id)
            
    n_x = num_nodes
    data.n_x = n_x

    data.train_percent = train_percent
    data.num_hyperedges = num_hyperedges
    
    return data

def load_yelp_dataset(path='../data/raw_data/yelp_raw_datasets/', dataset = 'yelp', 
        name_dictionary_size = 1000,
        train_percent = 0.025):

    latlong = pd.read_csv(osp.join(path, 'yelp_restaurant_latlong.csv')).values

    loc = pd.read_csv(osp.join(path, 'yelp_restaurant_locations.csv'))
    state_int = loc.state_int.values
    city_int = loc.city_int.values

    num_nodes = loc.shape[0]
    state_1hot = np.zeros((num_nodes, state_int.max()))
    state_1hot[np.arange(num_nodes), state_int - 1] = 1

    city_1hot = np.zeros((num_nodes, city_int.max()))
    city_1hot[np.arange(num_nodes), city_int - 1] = 1

    vectorizer = CountVectorizer(max_features = name_dictionary_size, stop_words = 'english', strip_accents = 'ascii')
    res_name = pd.read_csv(osp.join(path, 'yelp_restaurant_name.csv')).values.flatten()
    name_bow = vectorizer.fit_transform(res_name).todense()

    features = np.hstack([latlong, state_1hot, city_1hot, name_bow])

    df_labels = pd.read_csv(osp.join(path, 'yelp_restaurant_business_stars.csv'))
    labels = df_labels.values.flatten()

    num_nodes, feature_dim = features.shape
    assert num_nodes == len(labels)
    print(f'number of nodes:{num_nodes}, feature dimension: {feature_dim}')

    features = torch.FloatTensor(features)
    labels = torch.LongTensor(labels)

    H = pd.read_csv(osp.join(path, 'yelp_restaurant_incidence_H.csv'))
    node_list = H.node.values - 1
    edge_list = H.he.values - 1 + num_nodes

    edge_index = np.vstack([node_list, edge_list])
    edge_index = np.hstack([edge_index, edge_index[::-1, :]])

    edge_index = torch.LongTensor(edge_index)

    data = Data(x = features,
                edge_index = edge_index,
                y = labels)

    total_num_node_id_he_id = edge_index.max() + 1
    data.edge_index, data.edge_attr = coalesce(data.edge_index, 
            None, 
            total_num_node_id_he_id, 
            total_num_node_id_he_id)
            
    n_x = num_nodes
    data.n_x = n_x
    
    data.train_percent = train_percent
    data.num_hyperedges = H.he.values.max()
    
    return data

def load_cornell_dataset(path='../data/raw_data/', dataset = 'amazon', 
        feature_noise = 0.1,
        feature_dim = None,
        train_percent = 0.025):

    # first load node labels
    df_labels = pd.read_csv(osp.join(path, dataset, f'node-labels-{dataset}.txt'), names = ['node_label'])
    num_nodes = df_labels.shape[0]
    labels = df_labels.values.flatten()

    # then create node features.
    num_classes = df_labels.values.max()
    features = np.zeros((num_nodes, num_classes))

    features[np.arange(num_nodes), labels - 1] = 1
    if feature_dim is not None:
        num_row, num_col = features.shape
        zero_col = np.zeros((num_row, feature_dim - num_col), dtype = features.dtype)
        features = np.hstack((features, zero_col))

    features = np.random.normal(features, feature_noise, features.shape)
    print(f'number of nodes:{num_nodes}, feature dimension: {features.shape[1]}')

    features = torch.FloatTensor(features)
    labels = torch.LongTensor(labels)

    p2hyperedge_list = osp.join(path, dataset, f'hyperedges-{dataset}.txt')
    node_list = []
    he_list = []
    he_id = num_nodes

    with open(p2hyperedge_list, 'r') as f:
        for line in f:
            if line[-1] == '\n':
                line = line[:-1]
            cur_set = line.split(',')
            cur_set = [int(x) for x in cur_set]

            node_list += cur_set
            he_list += [he_id] * len(cur_set)
            he_id += 1
    # shift node_idx to start with 0.
    node_idx_min = np.min(node_list)
    node_list = [x - node_idx_min for x in node_list]

    edge_index = [node_list + he_list, 
                  he_list + node_list]

    edge_index = torch.LongTensor(edge_index)

    data = Data(x = features,
                edge_index = edge_index,
                y = labels)

    total_num_node_id_he_id = edge_index.max() + 1
    data.edge_index, data.edge_attr = coalesce(data.edge_index, 
            None, 
            total_num_node_id_he_id, 
            total_num_node_id_he_id)
            
    n_x = num_nodes
    data.n_x = n_x
    
    data.train_percent = train_percent
    data.num_hyperedges = he_id - num_nodes
    
    return data

def load_LE_dataset(path=None, dataset="ModelNet40", train_percent = 0.025):
    # load edges, features, and labels.
    print('Loading {} dataset...'.format(dataset))
    
    file_name = f'{dataset}.content'
    p2idx_features_labels = osp.join(path, dataset, file_name)
    idx_features_labels = np.genfromtxt(p2idx_features_labels,
                                        dtype=np.dtype(str))
    features = sp.csr_matrix(idx_features_labels[:, 1:-1], dtype=np.float32)
    labels = torch.LongTensor(idx_features_labels[:, -1].astype(float))

    print ('load features')
    
    # build graph
    idx = np.array(idx_features_labels[:, 0], dtype=np.int32)
    idx_map = {j: i for i, j in enumerate(idx)}
    
    file_name = f'{dataset}.edges'
    p2edges_unordered = osp.join(path, dataset, file_name)
    edges_unordered = np.genfromtxt(p2edges_unordered,
                                    dtype=np.int32)
    
    edges = np.array(list(map(idx_map.get, edges_unordered.flatten())),
                     dtype=np.int32).reshape(edges_unordered.shape)

    print ('load edges')

    # From adjacency matrix to edge_list
    edge_index = edges.T 
    assert edge_index[0].max() == edge_index[1].min() - 1

    # check if values in edge_index is consecutive. i.e. no missing value for node_id/he_id.
    assert len(np.unique(edge_index)) == edge_index.max() + 1
    
    num_nodes = edge_index[0].max() + 1
    num_he = edge_index[1].max() - num_nodes + 1
    
    edge_index = np.hstack((edge_index, edge_index[::-1, :]))
    
    # build torch data class
    data = Data(
            x = torch.FloatTensor(np.array(features[:num_nodes].todense())), 
            edge_index = torch.LongTensor(edge_index),
            y = labels[:num_nodes])

    total_num_node_id_he_id = len(np.unique(edge_index))
    data.edge_index, data.edge_attr = coalesce(data.edge_index, 
            None, 
            total_num_node_id_he_id, 
            total_num_node_id_he_id)
            
    n_x = num_nodes
    num_class = len(np.unique(labels[:num_nodes].numpy()))
    data.n_x = n_x
    data.y=data.y-data.y.min()
    
    data.train_percent = train_percent
    data.num_hyperedges = num_he
    
    return data



import os
import random
import torch
import numpy as np
from dhgbench.lib_utils.utils import fix_seed
from dhgbench.lib_dataset.edge_sampler import *

def generate_split_hyperedges(data,args,seed):

    hyperedge_index = data.hyperedge_index.to('cpu')

    hyperedges = defaultdict(set)

    for node, edge in zip(*hyperedge_index.tolist()):
        hyperedges[edge].add(node)

    HE = [frozenset(nodes) for nodes in hyperedges.values()]
    
    base_cover = get_cover_idx(HE)
    union = get_union(HE)
    tmp = [HE[idx] for idx in base_cover]
    assert union == get_union(tmp)
    base_num = len(base_cover)
    
    os.makedirs(args.edge_save_dir+args.dname, exist_ok=True)  
    
    seed_base = 42  
    fix_seed(seed_base+seed) 
    
    # ground 60%, train 10(+50)%, validation 10(+10)%, test 20%
    # ground_num = int(0.6*len(HE)) - base_num
    ground_num = max(int(0.6*len(HE)) - base_num,0)
    total_idx = list(range(len(HE))) 
    ground_idx = list(set(total_idx)-set(base_cover))
    ground_idx = random.sample(ground_idx, ground_num)      
    ground_num += base_num
    ground_idx += base_cover
    ground_valid_num = ground_num//6
    ground_valid_idx = random.sample(ground_idx, ground_valid_num)
    ground_train_num = ground_num - ground_valid_num
    
    ground_train_data = []
    ground_valid_data = []
    pred_data = []
    for idx in total_idx :
        if idx in ground_idx:
            if idx in ground_valid_idx:
                ground_valid_data.append(HE[idx])
            else:
                ground_train_data.append(HE[idx])
        else :
            pred_data.append(HE[idx])
            
    valid_only_num = int(0.25*len(pred_data))
    train_only_num = int(0.25*len(pred_data))
    test_num = len(pred_data) - (valid_only_num + train_only_num)
    
    random.shuffle(pred_data)
    train_only_data = pred_data[:train_only_num]
    valid_only_data = pred_data[train_only_num:-test_num]
    test_data = pred_data[-test_num:]
    
    # negative sampling        
    GP_train = ground_valid_data + ground_train_data + train_only_data
    GP_valid = ground_valid_data + ground_train_data + train_only_data + valid_only_data
    GP_test = GP_valid
    
    train_mns, train_sns, train_cns = neg_generator(GP_train, ground_train_num+train_only_num)
    valid_mns, valid_sns, valid_cns = neg_generator(GP_valid, ground_valid_num+valid_only_num)
    test_mns, test_sns, test_cns = neg_generator(GP_test, test_num)
    
    # positive samples
    ground_train_data = [list(edge) for edge in ground_train_data]
    ground_valid_data = [list(edge) for edge in ground_valid_data]
    train_only_data = [list(edge) for edge in train_only_data]
    valid_only_data = [list(edge) for edge in valid_only_data]
    test_data = [list(edge) for edge in test_data]
    
    print(f"ground {len(ground_train_data)} + {len(ground_valid_data)} = {len(ground_train_data + ground_valid_data)}")
    print(f"train pos {len(ground_train_data)} + {len(train_only_data)} = {len(ground_train_data + train_only_data)}, neg {len(train_mns)}")
    print(f"valid pos {len(ground_valid_data)} + {len(valid_only_data)} = {len(ground_valid_data + valid_only_data)}, neg {len(valid_mns)}")
    print(f"test pos {len(test_data)}, neg {len(test_mns)}")
    
    torch.save({'ground_train': ground_train_data, 'ground_valid': ground_valid_data, \
        'train_only_pos': train_only_data, 'train_mns': train_mns, 'train_sns' : train_sns, 'train_cns' : train_cns,\
        'valid_only_pos': valid_only_data, 'valid_mns': valid_mns, 'valid_sns' : valid_sns, 'valid_cns' : valid_cns, \
        'test_pos': test_data, 'test_mns': test_mns, 'test_sns' : test_sns, 'test_cns' : test_cns},
        f'./lib_edge_splits/{args.dname}/split_{seed}.pt')

def generate_edge_loaders(data_dict, args):

    device = args.device
    
    train_pos_loader = load_train(data_dict, args.edge_batch_size, device,label="pos") # only positives
    train_neg_loader = load_train(data_dict, args.edge_batch_size, device,label=args.ns_method) # only positives

    val_pos_loader = load_val(data_dict, args.edge_batch_size, device, label="pos")
    val_neg_sns_loader = load_val(data_dict, args.edge_batch_size, device, label="sns")
    val_neg_mns_loader = load_val(data_dict, args.edge_batch_size, device, label="mns")
    val_neg_cns_loader = load_val(data_dict, args.edge_batch_size, device, label="cns")

    test_pos_loader = load_test(data_dict, args.edge_batch_size, device, label="pos")
    test_neg_sns_loader = load_test(data_dict, args.edge_batch_size, device, label="sns")
    test_neg_mns_loader = load_test(data_dict, args.edge_batch_size, device, label="mns")
    test_neg_cns_loader = load_test(data_dict, args.edge_batch_size, device, label="cns")

    batch_loaders = {
        'train_pos': train_pos_loader,
        'train_neg': train_neg_loader,
        'val_pos': val_pos_loader,
        'val_neg_sns':val_neg_sns_loader,
        'val_neg_mns':val_neg_mns_loader,
        'val_neg_cns':val_neg_cns_loader,
        'test_pos': test_pos_loader,
        'test_neg_sns':test_neg_sns_loader,
        'test_neg_mns':test_neg_mns_loader,
        'test_neg_cns':test_neg_cns_loader,
    }
    
    return batch_loaders

def load_train(data_dict, bs, device, label):
    if label=="pos":
        train_pos = data_dict["train_only_pos"] + data_dict["ground_train"]
        train_pos_label = [1 for i in range(len(train_pos))]
        train_batchloader = HEBatchGenerator(train_pos, train_pos_label, bs, device, test_generator=False) 
    elif label =="mixed":
        d = len(data_dict["train_sns"]) // 3
        train_neg = data_dict["train_sns"][0:d] + data_dict["train_mns"][0:d] + data_dict["train_cns"][0:d]
        train_neg_label = [0 for i in range(len(train_neg))]
        train_batchloader = HEBatchGenerator(train_neg, train_neg_label, bs, device, test_generator=False) 
    else:
        train_neg = data_dict[f"train_{label}"]
        train_neg_label = [0 for i in range(len(train_neg))]
        train_batchloader = HEBatchGenerator(train_neg, train_neg_label, bs, device, test_generator=False) 
    
    return train_batchloader

def load_val(data_dict, bs, device, label):
    if label=="pos":
        val = data_dict["train_only_pos"] + data_dict["ground_train"]
        val_label = [1 for i in range(len(val))]
    else:
        val = data_dict[f"valid_{label}"]
        val_label = [0 for i in range(len(val))]
    val_batchloader = HEBatchGenerator(val, val_label, bs, device, test_generator=True)    
    return val_batchloader

def load_test(data_dict, bs, device, label):
    test = data_dict[f"test_{label}"]
    if label=="pos":
        test_label = [1 for i in range(len(test))]
    else:
        test_label = [0 for i in range(len(test))]
    test_batchloader = HEBatchGenerator(test, test_label, bs, device, test_generator=True)    
    return test_batchloader

# Batch Generator
class HEBatchGenerator(object):
    def __init__(self, hyperedges, labels, batch_size, device, test_generator=False):
        """Creates an instance of HyperedgeGroupBatchGenerator.
        
        Args:
            hyperedges: List(frozenset). List of hyperedges.
            labels: list. Labels of hyperedges.
            batch_size. int. Batch size of each batch.
            test_generator: bool. Whether batch generator is test generator.
        """
        self.batch_size = batch_size
        self.hyperedges = hyperedges
        self.labels = labels
        self._cursor = 0
        self.device = device
        self.test_generator = test_generator
        self.shuffle()
    
    def eval(self):
        self.test_generator = True

    def train(self):
        self.test_generator = False

    def shuffle(self):
        idcs = np.arange(len(self.hyperedges))
        np.random.shuffle(idcs)
        self.hyperedges = [self.hyperedges[i] for i in idcs]
        self.labels = [self.labels[i] for i in idcs]
  
    def __iter__(self):
        self._cursor = 0
        return self
    
    def next(self):
        return self.__next__()

    def __next__(self):
        if self.test_generator:
            return self.next_test_batch()
        else:
            return self.next_train_batch()

    def next_train_batch(self):
        ncursor = self._cursor+self.batch_size
        if ncursor >= len(self.hyperedges):
            hyperedges = self.hyperedges[self._cursor:] + self.hyperedges[
                :ncursor - len(self.hyperedges)]

            labels = self.labels[self._cursor:] + self.labels[
                :ncursor - len(self.labels)]
          
            self._cursor = ncursor - len(self.hyperedges)
            hyperedges = [torch.LongTensor(edge).to(self.device) for edge in hyperedges]
            labels = torch.FloatTensor(labels).to(self.device)
            self.shuffle()
            return hyperedges, labels, True
        
        hyperedges = self.hyperedges[
            self._cursor:self._cursor + self.batch_size]
        
        labels = self.labels[
            self._cursor:self._cursor + self.batch_size]
        
        hyperedges = [torch.LongTensor(edge).to(self.device) for edge in hyperedges]
        labels = torch.FloatTensor(labels).to(self.device)
       
        self._cursor = ncursor % len(self.hyperedges)
        return hyperedges, labels, False

    def next_test_batch(self):
        ncursor = self._cursor+self.batch_size
        if ncursor >= len(self.hyperedges):
            hyperedges = self.hyperedges[self._cursor:]
            labels = self.labels[self._cursor:]
            self._cursor = 0
            hyperedges = [torch.LongTensor(edge).to(self.device) for edge in hyperedges]
            labels = torch.FloatTensor(labels).to(self.device)
            
            return hyperedges, labels, True
        
        hyperedges = self.hyperedges[
            self._cursor:self._cursor + self.batch_size]
        
        labels = self.labels[
            self._cursor:self._cursor + self.batch_size]

        hyperedges = [torch.LongTensor(edge).to(self.device) for edge in hyperedges]
        labels = torch.FloatTensor(labels).to(self.device)
       
        self._cursor = ncursor % len(self.hyperedges)
        return hyperedges, labels, False

def get_union(union):
    ind = []
    for s in union :
        ind+=list(s)
    return set(ind)

def set_cover(universe, subsets):
    elements = set(e for s in subsets for e in s)
    if elements != universe:
        return None, None
    covered = set()
    cover = []
    idx = []
    while covered != elements:
        subset = max(subsets, key=lambda s: len(s - covered))
        cover.append(subset)
        idx.append(subsets.index(subset))
        covered |= subset
    return cover, idx

def get_cover_idx(HE):
    universe = get_union(HE)
    tmp_HE = [set(edge) for edge in HE]
    _, cover_idx = set_cover(universe, tmp_HE)
    return cover_idx



import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
import math
import torch_scatter
from dhgbench.lib_models.HNN.utils import normalize_l2

class UniGCNIIConv(nn.Module):
    def __init__(self, args, in_features, out_features):
        super().__init__()
        self.in_features, self.out_features = in_features, out_features
        self.W = nn.Linear(in_features, out_features, bias=False)
        self.args = args
        self.reduce = args.reduce

    def reset_parameters(self):
        self.W.reset_parameters()
        
    def forward(self, X, vertex, edges, alpha, beta, X0):
        N = X.shape[-2]

        Xve = X[..., vertex, :] # [nnz, C]
        Xe = torch_scatter.scatter(Xve, edges, dim=-2, reduce=self.reduce) # [E, C], reduce is 'mean' here as default

        Xev = Xe[..., edges, :] # [nnz, C]
        Xv = torch_scatter.scatter(Xev, vertex, dim=-2, reduce=self.reduce, dim_size=N) # [N, C]

        X = Xv 

        if self.args.use_norm:
            X = normalize_l2(X)

        Xi = (1-alpha) * X + alpha * X0
        X = (1-beta) * Xi + beta * self.W(Xi)

        return X, Xe

    def flops(self, X, vertex, edges):
        flops = vertex.shape[0] + edges.shape[0] + edges.shape[0] + vertex.shape[0] # scatter
        flops += np.prod(X.shape) # init connection
        flops += np.prod(X.shape[:-1]) * self.in_features * self.out_features # linear
        return flops


class UniGCNII(nn.Module):
    def __init__(self, num_features, num_targets, args):

        super().__init__()
        self.num_layers = args.All_num_layers
        self.nhid = args.MLP_hidden

        self.in_features = num_features
        self.out_features = num_targets

        act = {'relu': nn.ReLU(), 'prelu':nn.PReLU() }
        self.act = act[args.activation] # Default relu
        self.input_drop = nn.Dropout(args.input_drop) # 0.6 is chosen as default
        self.dropout = nn.Dropout(args.dropout) # 0.2 is chosen for GCNII
        self.alpha = args.restart_alpha
        self.lamda = args.lamda

        self.convs = torch.nn.ModuleList()
        self.convs.append(torch.nn.Linear(num_features, self.nhid))
        for _ in range(self.num_layers):
            self.convs.append(UniGCNIIConv(args, self.nhid, self.nhid))
        self.convs.append(torch.nn.Linear(self.nhid, num_targets))
        self.reg_params = list(self.convs[1:-1].parameters())
        self.non_reg_params = list(self.convs[0:1].parameters())+list(self.convs[-1:].parameters())

    def reset_parameters(self):
        for conv in self.convs:
            conv.reset_parameters()
        
    def forward(self, data):
        x = data.x
        V, E = data.hyperedge_index[0], data.hyperedge_index[1]
        lamda, alpha = self.lamda, self.alpha
        x = self.dropout(x)
        x = F.relu(self.convs[0](x)) 
        x0 = x 
        for i,con in enumerate(self.convs[1:-1]):
            x = self.dropout(x)
            beta = math.log(lamda/(i+1)+1)
            x,e = con(x, V, E, alpha, beta, x0)
            x = F.relu(x)
        x = self.dropout(x)
        x = self.convs[-1](x)
        return x,e

    def flops(self, data):
        x = data.x
        V, E = data.edge_index[0], data.edge_index[1]
        flops = self.in_features * self.nhid * np.prod(x.shape[:-1]) # linear
        flops += self.nhid * np.prod(x.shape[:-1]) # non-linear
        for i,con in enumerate(self.convs[1:-1]):
            flops += con.flops(x, V, E) # conv
            flops += self.nhid * np.prod(x.shape[:-1]) # non-linear
        flops = self.out_features * self.nhid * np.prod(x.shape[:-1]) # linear
        return flops


import math
import torch
import torch.nn as nn
import torch.nn.functional as F
from dhgbench.lib_models.HNN.utils import SparseMM

class HJRLConv(nn.Module):

    def __init__(self, in_features, out_features, activation='leaky_relu', neg_slope=0, bias=False):
        super(HJRLConv, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.activation = activation

        act = {'relu': F.relu, 'elu': F.elu, 'leaky_relu': F.leaky_relu}
        if self.activation is not None:
            self.act = act[activation]

        self.neg_slope = neg_slope
        self.weight = nn.Parameter(torch.FloatTensor(in_features, out_features))
        if bias:
            self.bias = nn.Parameter(torch.FloatTensor(out_features))
        else:
            self.register_parameter('bias', None)

    def reset_parameters(self):
        stdv = 1. / math.sqrt(self.weight.size(1))
        self.weight.data.uniform_(-stdv, stdv)
        if self.bias is not None:
            self.bias.data.uniform_(-stdv, stdv)

    def forward(self, input_features, adj):
        support = SparseMM.apply(input_features, self.weight)
        output = SparseMM.apply(adj, support)

        if self.bias is not None:
            output = output + self.bias
        elif self.activation == None:
            output = output
        elif self.activation == 'leaky_relu':
            output = self.act(output, negative_slope=self.neg_slope)
        else:
            output = self.act(output)
        
        return output

    def __repr__(self):
        return self.__class__.__name__ + ' (' \
               + str(self.in_features) + ' -> ' \
               + str(self.out_features) + ')'

class HJRL(nn.Module):
    def __init__(self, num_features, num_targets, args):
        super(HJRL, self).__init__()

        self.num_layers = args.All_num_layers
        self.hidden_dim =args.MLP_hidden
        self.dropout = args.dropout  # Note that default is 0.6
        self.activition = args.activation
        self.neg_slope = args.neg_slope

        self.convs = nn.ModuleList()
        if self.num_layers == 1:
            self.convs.append(HJRLConv(num_features,num_targets,activation=None))
        else:
            self.convs.append(HJRLConv(num_features,self.hidden_dim,self.activition,self.neg_slope))
            for _ in range(self.num_layers-2):
                self.convs.append(HJRLConv(self.hidden_dim,self.hidden_dim,self.activition,self.neg_slope))
            # Output heads is set to 1 as default
            self.convs.append(HJRLConv(
                self.hidden_dim, num_targets,activation=None))

    def reset_parameters(self):
        for layer in self.convs:
            layer.reset_parameters()

    def forward(self,data):

        x, y, HHT, H, HT, HTH = data.norm_X, data.norm_E, data.HHT, data.H, data.HT, data.HTH
        
        if self.num_layers == 1:
            HHT_l = self.convs[0](x,HHT)
            H_l = self.convs[0](y, H)
            HT_l = self.convs[0](x, HT)
            HTH_l = self.convs[0](y, HTH)
        else:
            for i, conv in enumerate(self.convs[:-1]):
                HHT_l = self.convs[i](x,HHT)
                H_l = self.convs[i](y, H)
                HT_l = self.convs[i](x, HT)
                HTH_l = self.convs[i](y, HTH)

                HHT_l = F.dropout(HHT_l, self.dropout, training=self.training)
                H_l = F.dropout(H_l, self.dropout, training=self.training)
                HT_l = F.dropout(HT_l, self.dropout, training=self.training)
                HTH_l = F.dropout(HTH_l, self.dropout, training=self.training)

                x = HHT_l + H_l
                y = HT_l + HTH_l

            HHT_l = self.convs[-1](x,HHT)
            H_l = self.convs[-1](y, H)
            HT_l = self.convs[-1](x, HT)
            HTH_l = self.convs[-1](y, HTH)

        HHT_l = F.leaky_relu(HHT_l, negative_slope=self.neg_slope)
        H_l = F.leaky_relu(H_l, negative_slope=self.neg_slope)
        HT_l = F.leaky_relu(HT_l, negative_slope=self.neg_slope)
        HTH_l = F.leaky_relu(HTH_l, negative_slope=self.neg_slope)

        x = HHT_l + H_l
        y = HT_l + HTH_l

        return x, y


from collections import defaultdict
from dhgbench.lib_utils.metrics import *
from dhgbench.lib_models import _semi_methods_
from dhgbench.lib_dataset import _fair_datasets_

class Evaluator:
    
    def __init__(self,args, **kwargs):
        """
        Training pipline for different kinds of models
        """
        self.args = args
        self.device=args.device
        
    @torch.no_grad()
    def node_cls_evaluation(self, model, data, masks, result=None):

        if result is not None:
            logits = result
        else:
            model.eval()
            logits,_ = model(data) 

        accs = accuracy(logits, data.y, masks)
        metrics_dict = {'acc': accs}

        return metrics_dict
    
    @torch.no_grad()
    def edge_pred_evaluation(self, model, data, batch_loaders, result=None):
        train_metrics, val_metrics, test_metrics = evaluate_edge(model,data,batch_loaders)
        return train_metrics, val_metrics, test_metrics
    
    @torch.no_grad()
    def hg_cls_evaluation(self, model, batch_loaders, result=None):
        result = evaluate_hypegraph(model,batch_loaders,self.args)
        return result
    
    @torch.no_grad()
    def node_cls_with_fairness(self, model, data, masks, result=None):
        
        if result is not None:
            logits = result
        else:
            model.eval()
            logits,_ = model(data) # logits
        
        accs = accuracy(logits, data.y, masks)
        
        f1s = f1_scores(logits,data.y,masks)
        
        aucs = auc_rocs(logits,data.y,masks)
        
        dps,eos = fairness(logits,data.y,masks,data.sens)

        metrics_dict={'acc':accs,'f1_score':f1s,'auc':aucs,'parity':dps,'equality':eos}
        return metrics_dict

    def evaluate(self,model,data,seed_split=None,task_type=None,verbose=False):
        
        metrics_dict=defaultdict(list) 
        
        if self.args.method in _semi_methods_:
            
            if task_type == 'node_cls':
                
                if data.name in _fair_datasets_:
                    metrics=self.node_cls_with_fairness(model, data, seed_split)
                else:
                    metrics=self.node_cls_evaluation(model, data, seed_split)
                    
                if verbose:
                    for m in metrics:
                        print(f'train_{m}: {metrics[m][0]:.2f}, valid_{m}: {metrics[m][1]:.2f}, test_{m}: {metrics[m][2]:.2f} ')
                
                metrics_dict = metrics_dict |  metrics     

            elif task_type == 'edge_pred':
                
                metrics_dict = defaultdict(None)
                train_metrics, val_metrics, test_metrics = self.edge_pred_evaluation(model,data,seed_split)
                
                if verbose:
                    edge_evaluation_printer(train_metrics, val_metrics, test_metrics)
                
                metrics_dict['train'],metrics_dict['val'],metrics_dict['test'] = train_metrics, val_metrics, test_metrics
            
            elif task_type == 'hg_cls':
                
                result = self.hg_cls_evaluation(model,seed_split)
                
                if verbose:
                    hg_evaluation_printer(result)
                
                metrics_dict = result 
            
        else:
            raise NotImplementedError
        
        return metrics_dict
    
 


import math
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch_scatter
from torch_scatter import scatter_softmax, scatter_sum, scatter_mean, scatter_add
from dhgbench.lib_models.HNN.utils import SparseMM

class DPHGNN(nn.Module):
    def __init__(self, num_features, num_targets, args):
        super(DPHGNN, self).__init__()
        
        self.fc_dim = args.fc_dim
        self.dff_MLP_hidden = args.dff_MLP_hidden
        
        self.taa_layer = TAAMSG(num_features,args)
        self.spectral_layer = HGSpectralNet(num_features,args.spectral_embed_dim)
        self.fc = nn.Linear(in_features=self.spectral_layer.out_channels+self.taa_layer.out_channels, 
                            out_features=self.fc_dim).to(args.device)
        self.dff_layer = DFF(self.fc_dim,num_targets,args)
    
    def reset_parameters(self):
        self.taa_layer.reset_parameters()
        self.spectral_layer.reset_parameters()
        self.fc.reset_parameters()
        self.dff_layer.reset_parameters()
        
    def forward(self,data):
        
        # 1. TAA message passing
        taa_features,_, S_star = self.taa_layer(data)
        
        # 2. feature mixture module
        spectral_features = self.spectral_layer(data)
        
        X = torch.cat([taa_features, spectral_features], dim=1) 
        X = self.fc(X)
        
        # 3. Dynamic Feature Fusion
        hyperedge_index = data.hyperedge_index
        X = self.dff_layer(X,hyperedge_index,S_star)
        
        return X,None

#--------------------- Dynamic Feature Fusion Module ---------------------

class DFF(nn.Module):
    def __init__(self, in_channels:int,out_channels: int, args):
        super(DFF, self).__init__()
        
        self.dff_MLP_hidden=args.dff_MLP_hidden 
        self.dff_num_layers = args.dff_num_layers
        
        self.convs = nn.ModuleList()
        self.convs.append(DPHGNNConv(in_channels, self.dff_MLP_hidden,args,is_last=False))
        for _ in range(self.dff_num_layers-1):
            self.convs.append(DPHGNNConv(self.dff_MLP_hidden, self.dff_MLP_hidden,args,is_last=False))
        self.convs.append(HypergraphConv(self.dff_MLP_hidden, out_channels, is_last=True))

    def reset_parameters(self):
        for conv in self.convs:
            conv.reset_parameters()
    
    def forward(self, X, HG, S_features):
        for conv in self.convs:
            X = conv(X, HG, S_features)
        return X

class HypergraphConv(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels, 
        bias: bool = True,
        drop_rate: float = 0.5,
        is_last: bool = False,
    ):
        super().__init__()
        self.is_last = is_last
        self.dropout=drop_rate
        self.W = nn.Linear(in_channels, out_channels, bias=bias)

    def reset_parameters(self):
        self.W.reset_parameters()
    
    def forward(self, X, hyperedge_index, S_features):
        
        V, E = hyperedge_index
        num_nodes,num_edges = V.max()+1,E.max()+1
        
        X = self.W(X)
        
        # Y = hg.v2e(X, aggr="mean")
        Xve = X[V]  # Shape: (num_edges, feature_dim)
        Y = scatter_mean(Xve, E, dim=0, dim_size=num_edges)  # Shape: (num_edges, feature_dim)
        
        ones = torch.ones(V.shape[0], device=V.device, dtype=torch.float32)
        D_v = torch_scatter.scatter_add(ones, V, dim=0)
        
        _De = torch.zeros(num_edges, device=V.device)
        _De = _De.scatter_reduce(0, index=E, src=D_v[V], reduce="mean")
        _De = _De.pow(-0.5)
        _De[_De.isinf()] = 1
        Y = _De.view(-1, 1) * Y
        
        Xev = Y[E]
        X = scatter_add(Xev, V, dim=0, dim_size=num_nodes)
        _Dv = D_v.pow(-0.5)
        _Dv[_Dv.isinf()] = 0
        X = _Dv.view(-1, 1) * X
        
        if not self.is_last:
            X = F.relu(X)
            X = F.dropout(X, self.dropout, training=self.training)
            
        return X

class DPHGNNConv(nn.Module):
    def __init__(self,
               in_channels,
               out_channels,
               args,
               bias: bool=True,
               drop_rate: float=0.5,
               is_last: bool=False
               ):
        super(DPHGNNConv, self).__init__()
        
        self.is_last = is_last
        self.star_dim = args.expan_dim
        
        self.W_x = nn.Linear(in_channels, out_channels, bias=bias)
        self.W_vertex = nn.Linear(in_channels, out_channels, bias=bias)
        self.atten_vertex = nn.Linear(out_channels, 1, bias=False) 
        
        self.theta_e2v = nn.Linear(out_channels+self.star_dim, out_channels, bias = bias)
        
        self.atten_act = nn.LeakyReLU(args.atten_neg_slope)
        self.dropout=drop_rate
    
    def reset_parameters(self):
        self.W_x.reset_parameters()
        self.W_vertex.reset_parameters()
        self.atten_vertex.reset_parameters()
        self.theta_e2v.reset_parameters()
    
    def forward(self, X, hyperedge_index, S_features):
        
        V, E = hyperedge_index
        num_nodes,num_edges = V.max()+1,E.max()+1
        
        X_init = self.W_x(X) 
        
        X_feat = self.W_vertex(X) # [|V|,out_channels]
        x_for_vertex = self.atten_vertex(X_feat) # [|V|,1]
        
        v2e_atten_score = x_for_vertex[hyperedge_index[0]]
        v2e_atten_score = F.dropout(self.atten_act(v2e_atten_score).squeeze(),self.dropout, training=self.training)

        softmax_weight = scatter_softmax(v2e_atten_score, E, dim=0)  # (nnz,)

        Xev = X_feat[V] * softmax_weight.unsqueeze(-1)  # (nnz, d)
        Y_v2e = scatter_sum(Xev, E, dim=0, dim_size=num_edges)  # (num_edges, d)
        Y_v2e = F.elu(Y_v2e)
        
        e_msg = torch.cat((Y_v2e, S_features), dim=1) 
        Y = self.theta_e2v(e_msg)
        
        Yev = Y[E]  # (nnz, d)

        X = scatter_mean(Yev, V, dim=0, dim_size=num_nodes)  # (num_nodes, d)
        
        if not self.is_last:
            X = F.elu(X)
            X = F.dropout(X, self.dropout, training=self.training)
            
        return X + X_init 
    
#--------------------- Topology-Aware Attention Layer ---------------------      
    
class TAAMSG(nn.Module):
    def __init__(self, num_features, args):
        super(TAAMSG, self).__init__()
        
        self.expan_dim = args.expan_dim
        self.spatial_feat_dim =args.taa_spatial_dim
        self.spectral_feat_dim = args.taa_spectral_dim
        self.num_heads =args.num_heads 
        self.chunk_size = args.chunk_size 
        self.out_channels = self.spatial_feat_dim+self.spectral_feat_dim 
        
        self.cg_layer = ExpanNet(num_features,self.expan_dim,args,var='CG')
        self.hg_layer = ExpanNet(num_features,self.expan_dim,args,var='HG')
        self.sg_layer = ExpanNet(num_features,self.expan_dim,args,var='SG')
        self.taa_spatial_layer = TAA_Spatial(self.expan_dim,self.spatial_feat_dim,self.num_heads,self.chunk_size)
        self.spectral_net = SmoothNet(self.expan_dim,self.expan_dim)
        self.taa_spectral_layer = TAA_Spectral(self.expan_dim,self.spectral_feat_dim,self.num_heads,self.chunk_size)
    
    def reset_parameters(self):
        self.cg_layer.reset_parameters()
        self.hg_layer.reset_parameters()
        self.sg_layer.reset_parameters()
        self.taa_spatial_layer.reset_parameters()
        self.spectral_net.reset_parameters()
        self.taa_spectral_layer.reset_parameters()
    
    def get_star_features(self,data):
        x = data.x
        num_hyperedges = data.hyperedge_index[1].max()+1
        s_init = torch.zeros(num_hyperedges, data.num_features).to(x.device)
        x_star = torch.cat((x, s_init), 0)
        return x_star
    
    def forward(self,data):
        
        '''
        1. Expansion Graph Convolution
        '''
        x_star = self.get_star_features(data) 
        clique_graph_features = self.cg_layer(data.x,data.A_cg)
        hypergcn_graph_features = self.hg_layer(data.x,data.A_hyp) 
        star_graph_features = self.sg_layer(x_star,data.A_sg) # 10->64, [94318, 64]

        X_star, S_star = star_graph_features[data.v_mask], star_graph_features[~data.v_mask] # X_star: [66790, 64], S_star: [27528, 64]
        
        '''
        2. Feature Mixture Module
        '''
        X_hyp, X_cg, X_sg = self.spectral_net(hypergcn_graph_features,clique_graph_features,star_graph_features,data.L_hyp,data.L_cg,data.L_sg)
        X_spectral = self.taa_spectral_layer(X_hyp, X_cg, X_sg[data.v_mask]) # [66790, 10]
        
        # return hypergcn_graph_features, clique_graph_features, X_star
        
        X_spatial = self.taa_spatial_layer(hypergcn_graph_features, clique_graph_features, X_star) # [66790, 64]
        
        taa_features = torch.cat((X_spatial, X_spectral), 1) # [66790, 76]
        
        return taa_features, X_star, S_star    
    
class TAA_Spectral(nn.Module):
    def __init__(self, in_channels, out_channels, num_heads, chunk_size):
        super(TAA_Spectral, self).__init__()
        
        self.num_heads = num_heads
        self.chunk_size = chunk_size
        
        self.hyp_proj = nn.Linear(in_channels, out_channels)
        self.star_proj = nn.Linear(in_channels, out_channels)
        self.clique_proj = nn.Linear(in_channels, out_channels)
        self.MHA = nn.MultiheadAttention(out_channels, num_heads)

    def chunked_MHA(self, Q, K, V, chunk_size):
        num_nodes = Q.shape[0]
        outputs = []
        for i in range(0, num_nodes, chunk_size):
            Q_chunk = Q[i : i + chunk_size]
            K_chunk = K[i : i + chunk_size]
            V_chunk = V[i : i + chunk_size]
            attn_out, _ = self.MHA(Q_chunk, K_chunk, V_chunk)
            outputs.append(attn_out)
        return torch.cat(outputs, dim=0)

    def reset_parameters(self):
        
        self.hyp_proj.reset_parameters()
        self.star_proj.reset_parameters()
        self.clique_proj.reset_parameters()

        nn.init.xavier_uniform_(self.MHA.in_proj_weight)
        if self.MHA.in_proj_bias is not None:
            nn.init.constant_(self.MHA.in_proj_bias, 0.)

        nn.init.xavier_uniform_(self.MHA.out_proj.weight)
        if self.MHA.out_proj.bias is not None:
            nn.init.constant_(self.MHA.out_proj.bias, 0.)

    def forward(self, X_L_hyp, X_L_star, X_L_clique):
        
        hyp_proj = self.hyp_proj(X_L_hyp)
        star_proj = self.star_proj(X_L_star)
        clique_proj = self.clique_proj(X_L_clique)
        if self.chunk_size==-1:
            spectral_attn_output, _ = self.MHA(hyp_proj, star_proj, clique_proj)
        else:
            spectral_attn_output = self.chunked_MHA(hyp_proj, star_proj, clique_proj,self.chunk_size)
        
        return spectral_attn_output  

class SmoothNet(nn.Module):
    def __init__(self, in_channels, out_channels, bias=True):
        super(SmoothNet, self).__init__()
        self.W_hyp = nn.Linear(in_channels, out_channels, bias = bias)
        self.W_cg = nn.Linear(in_channels, out_channels, bias = bias)
        self.W_sg = nn.Linear(in_channels, out_channels, bias = bias)

    def reset_parameters(self):
        self.W_hyp.reset_parameters()
        self.W_cg.reset_parameters()
        self.W_sg.reset_parameters()
    
    def forward(self, X_HypGNet, X_CGNet, X_SGNet, L_hyp, L_cg, L_sg):
        
        X_hyp = self.W_hyp(X_HypGNet)
        X_cg = self.W_cg(X_CGNet)
        X_sg = self.W_sg(X_SGNet)
        
        X_hyp_s = SparseMM.apply(L_hyp, X_hyp)
        X_cg_s = SparseMM.apply(L_cg, X_cg)
        X_sg_s = SparseMM.apply(L_sg, X_sg)
        
        return X_hyp_s, X_cg_s, X_sg_s

class TAA_Spatial(nn.Module):
    def __init__(self, in_channels, out_channels, num_heads, chunk_size):
        super(TAA_Spatial, self).__init__()
        
        self.num_heads = num_heads
        self.chunk_size = chunk_size
        
        self.hyp_proj = nn.Linear(in_channels, out_channels)
        self.star_proj = nn.Linear(in_channels, out_channels)
        self.clique_proj = nn.Linear(in_channels, out_channels)
        self.MHA = nn.MultiheadAttention(out_channels, num_heads)

    def chunked_MHA(self, Q, K, V, chunk_size):
        num_nodes = Q.shape[0]
        outputs = []
        for i in range(0, num_nodes, chunk_size):
            Q_chunk = Q[i : i + chunk_size]
            K_chunk = K[i : i + chunk_size]
            V_chunk = V[i : i + chunk_size]
            attn_out, _ = self.MHA(Q_chunk, K_chunk, V_chunk)
            outputs.append(attn_out)
        return torch.cat(outputs, dim=0)
    
    def reset_parameters(self):
        
        self.hyp_proj.reset_parameters()
        self.star_proj.reset_parameters()
        self.clique_proj.reset_parameters()

        nn.init.xavier_uniform_(self.MHA.in_proj_weight)
        if self.MHA.in_proj_bias is not None:
            nn.init.constant_(self.MHA.in_proj_bias, 0.)

        nn.init.xavier_uniform_(self.MHA.out_proj.weight)
        if self.MHA.out_proj.bias is not None:
            nn.init.constant_(self.MHA.out_proj.bias, 0.)
    
    def forward(self, X_hyp, X_star, X_clique):
        
        hyp_proj = self.hyp_proj(X_hyp)
        star_proj = self.star_proj(X_star)
        clique_proj = self.clique_proj(X_clique)
        
        if self.chunk_size==-1:
            spatial_attn_output, _ = self.MHA(hyp_proj, star_proj, clique_proj)
        else:
            spatial_attn_output = self.chunked_MHA(hyp_proj, star_proj, clique_proj,self.chunk_size)
        
        return spatial_attn_output

class HGSpectralNet(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, bias=True, drop_rate: float=0.3):
        super(HGSpectralNet, self).__init__()
        self.dropout = drop_rate
        self.out_channels = out_channels
        self.W = nn.Linear(in_channels*2, out_channels, bias= bias)

    def reset_parameters(self):
        self.W.reset_parameters()
    
    def forward(self, data):
        
        x, L_hgnn, L_sym, L_rw = data.x,data.L_hgnn,data.L_sym,data.L_rw
        
        X_hgnn = SparseMM.apply(L_hgnn, x)
        X_sym = SparseMM.apply(L_sym, x)
        X_rw = SparseMM.apply(L_rw, x)
        
        X_sym_rw = (X_sym + X_rw)/2
        X_spectral = torch.cat([X_hgnn, X_sym_rw], dim=1)
        
        X = F.relu(self.W(X_spectral))
        X = F.dropout(X, self.dropout, training=self.training)

        return X

#--------------------- Convolution for Expansion Graph ---------------------

class ExpanNet(nn.Module):
    def __init__(self, in_channels, out_channels, args, var = 'CG'):
        super(ExpanNet, self).__init__()

        # var choices = ['CG','HG','SG']
        self.args =args
        self.num_layers = eval(f'args.{var.lower()}_num_layer')
        self.hidden_dim =eval(f'args.{var.lower()}_MLP_hidden')
        self.dropout = eval(f'args.{var.lower()}_dropout') 

        self.convs = nn.ModuleList()
        if self.num_layers == 1:
            self.convs.append(GCNConv(in_channels,out_channels))
        else:
            self.convs.append(GCNConv(in_channels,self.hidden_dim))
            for _ in range(self.num_layers-2):
                self.convs.append(GCNConv(self.hidden_dim,self.hidden_dim))
            # Output heads is set to 1 as default
            self.convs.append(GCNConv(
                self.hidden_dim, out_channels))

    def reset_parameters(self):
        for layer in self.convs:
            layer.reset_parameters()
            
    def forward(self,x,A):

        for i, conv in enumerate(self.convs[:-1]):
            x = conv(x,A)
            x = F.relu(x)
            x = F.dropout(x, p=self.dropout, training=self.training)

        x = self.convs[-1](x, A)
        
        return x

class GCNConv(nn.Module):

    def __init__(self, in_features, out_features, bias=True):
        super(GCNConv, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        
        self.weight = nn.Parameter(torch.FloatTensor(in_features, out_features))
        if bias:
            self.bias = nn.Parameter(torch.FloatTensor(out_features))
        else:
            self.register_parameter('bias', None)

    def reset_parameters(self):
        stdv = 1. / math.sqrt(self.weight.size(1))
        self.weight.data.uniform_(-stdv, stdv)
        if self.bias is not None:
            self.bias.data.uniform_(-stdv, stdv)

    def forward(self, input_features, adj):
        support = SparseMM.apply(input_features, self.weight)
        output = SparseMM.apply(adj, support)

        if self.bias is not None:
            output = output + self.bias
        else:
            output = output
        
        return output

    def __repr__(self):
        return self.__class__.__name__ + ' (' \
               + str(self.in_features) + ' -> ' \
               + str(self.out_features) + ')'



import math
import torch
import torch.nn as nn
from dhgbench.lib_models.HNN.mlp import MLP_inner

def util_sparse_einsum(abh: torch.sparse.Tensor, bhd: torch.Tensor) -> torch.Tensor:
    """abh,bhd->ahd"""
    a, b, h = abh.size()
    b_, h_, d = bhd.size()
    assert b == b_ and h == h_
    indices, values = abh.indices(), abh.values()  # [2, |indices|], [|indices|, h]
    # elementwise multiplication -> row summation
    res = torch.sparse_coo_tensor(
        indices[0:1], values[..., None] * bhd[indices[1]], size=(a, h, d)
    )  # [a, h, d]
    return res.coalesce().to_dense()  # [a, h, d]

class EHNNTransformerConv(nn.Module):
    def __init__(
        self,
        dim_in,
        dim_hidden,
        dim_qk,
        dim_ff,
        n_heads,
        dropout,
        hypernet_info,
        inner_dim,
        pe_dim,
        hyper_dim,
        hyper_layers,
        hyper_dropout,
        force_broadcast,
        input_dropout,
        mlp_classifier,
        att0_dropout,
        att1_dropout,
    ):
        super().__init__()
        assert mlp_classifier
        self.input_dropout = nn.Dropout(input_dropout)
        self.dropout = nn.Dropout(dropout, inplace=True)
        self.lin_proj = nn.Linear(dim_in, dim_hidden)
        
        self.v2e_layer = TransformerV2E(
            dim_hidden,
            dim_qk,
            dim_ff,
            n_heads,
            dropout,
            hypernet_info,
            inner_dim,
            pe_dim,
            hyper_dim,
            hyper_layers,
            hyper_dropout,
            force_broadcast,
            att0_dropout,
            att1_dropout,
        )
        self.e2v_layer = TransformerE2V(
            dim_hidden,
            dim_qk,
            dim_ff,
            n_heads,
            dropout,
            hypernet_info,
            inner_dim,
            pe_dim,
            hyper_dim,
            hyper_layers,
            hyper_dropout,
            force_broadcast,
            att0_dropout,
            att1_dropout,
        )

        print(f"hypernet max input range: {hypernet_info}")

    def reset_parameters(self):
        self.lin_proj.reset_parameters()
        self.v2e_layer.reset_parameters()
        self.e2v_layer.reset_parameters()

    def forward(self, x, ehnn_cache) -> torch.Tensor:
            
        x = self.input_dropout(x)
        x = self.lin_proj(x)

        x_v, x_e = self.v2e_layer(x, ehnn_cache)
        x_v, x_e = self.dropout(x_v), self.dropout(x_e)
        x = (x_v, x_e)
        x = self.e2v_layer(x, ehnn_cache)

        return x,x_e

class SelfAttnV2E(nn.Module):
    def __init__(
        self,
        dim_in,
        dim_qk,
        n_heads,
        hypernet_info,
        inner_dim,
        pe_dim,
        hyper_dim,
        hyper_layers,
        hyper_dropout,
        force_broadcast,
        att0_dropout,
        att1_dropout,
    ):
        super().__init__()
        _, max_l, _ = hypernet_info
        self.max_l = max_l
        self.dim_in = dim_in
        self.dim_qk = dim_qk
        self.n_heads = n_heads
        self.inner_dim = inner_dim
        self.dim_v = inner_dim
        self.pe_dim = pe_dim
        self.hyper_dim = hyper_dim
        self.hyper_layers = hyper_layers
        self.hyper_dropout = hyper_dropout
        self.force_broadcast = force_broadcast
        self.dim_qk_head = dim_qk // n_heads if dim_qk >= n_heads else 1
        self.dim_v_head = self.dim_v // n_heads if self.dim_v >= n_heads else 1
        self.q = PositionalMLP(
            n_heads * self.dim_qk_head,
            2,
            pe_dim,
            hyper_dim,
            hyper_layers,
            hyper_dropout,
        )
        self.k = nn.Linear(self.dim_v, 2 * dim_qk)
        self.v = nn.Linear(self.dim_v, self.dim_v)

        self.mlp1 = MLP_inner(dim_in, self.dim_v, hyper_dim, hyper_layers, hyper_dropout)
        self.mlp2 = MLP_inner(
            self.dim_v * 2, self.dim_v, hyper_dim, hyper_layers, hyper_dropout
        )
        self.mlp3 = MLP_inner(self.dim_v * 2, dim_in, hyper_dim, hyper_layers, hyper_dropout)
        self.norm1 = nn.LayerNorm(inner_dim)
        self.norm2 = nn.LayerNorm(inner_dim)
        self.norm3 = nn.LayerNorm(inner_dim)
        self.pe2 = PositionalEncoding(inner_dim, 2)
        self.pe3 = PositionalEncoding(inner_dim, max_l + 1)
        self.b = BiasE(
            dim_in,
            max_l,
            pe_dim,
            hyper_dim,
            hyper_layers,
            hyper_dropout,
            force_broadcast,
        )

        self.att0_dropout = nn.Dropout(att0_dropout)
        self.att1_dropout = nn.Dropout(att1_dropout)

    def reset_parameters(self):
        self.q.reset_parameters()
        self.k.reset_parameters()
        self.v.reset_parameters()
        self.mlp1.reset_parameters()
        self.mlp2.reset_parameters()
        self.mlp3.reset_parameters()
        self.norm1.reset_parameters()
        self.norm2.reset_parameters()
        self.norm3.reset_parameters()
        self.b.reset_parameters()

    def forward(self, x, ehnn_cache):
        """forward method
        :param x: [N, D]
        :param ehnn_cache:
        :return: [N, D], [|E|, D]
        """
        incidence = ehnn_cache["incidence"]
        edge_orders = ehnn_cache["edge_orders"]

        # MLP1
        x = x + self.mlp1(self.norm1(x))  # [N, D]

        q_0 = self.q(torch.zeros(1, dtype=torch.long, device=x.device)).view(
            self.n_heads, self.dim_qk_head
        )
        q_1 = self.q(torch.ones(1, dtype=torch.long, device=x.device)).view(
            self.n_heads, self.dim_qk_head
        )
        k = self.k(x)
        k_0, k_1 = k.split(self.dim_qk, -1)
        k_0 = torch.stack(k_0.split(self.dim_qk_head, -1), 1)  # [N, H, D/H]
        k_1 = torch.stack(k_1.split(self.dim_qk_head, -1), 1)  # [N, H, D/H]
        v = self.v(x)
        v = torch.stack(v.split(self.dim_v_head, -1), 1)  # [N, H, Dv/H]

        # aggregation
        # either 0-overlap (global), or 1-overlap (local)
        # 0-overlap
        logit0 = torch.einsum("hd,nhd->nh", q_0, k_0) / math.sqrt(
            self.dim_qk_head
        )  # [N, H]
        alpha0 = torch.softmax(logit0, dim=0)  # [N, H]
        att0 = torch.einsum("nh,nhd->hd", alpha0, v)  # [H, Dv/H]
        att0 = torch.cat(att0.unbind(-2), -1)  # [Dv,]
        att0 = att0[None, :]  # [1, Dv]
        # 1-overlap node-to-node
        att1_v = torch.cat(v.unbind(-2), -1)  # [N, Dv]
        # 1-overlap node-to-edge
        logit1 = torch.einsum("hd,nhd->nh", q_1, k_1) / math.sqrt(
            self.dim_qk_head
        )  # [N, H]
        # broadcast logits over incidence matrix
        n, e = incidence.size()
        indices = incidence.indices()[
            [1, 0]
        ]  # [2, |indices|], indices of transposed incidence matrix
        logit1 = torch.sparse_coo_tensor(
            indices, logit1[indices[1]], size=(e, n, self.n_heads)
        )  # [|E|, N, H]
        alpha1 = torch.sparse.softmax(logit1, dim=1).coalesce()  # [|E|, N, H]
        att1_e = util_sparse_einsum(alpha1, v)  # [|E|, H, Dv/H]
        att1_e = torch.cat(att1_e.unbind(-2), -1)  # [|E|, Dv]

        # MLP2
        pe2_s0 = self.pe2(torch.zeros(1, dtype=torch.long, device=x.device)).view(
            1, self.inner_dim
        )  # [1, D]
        pe2_s1 = self.pe2(torch.ones(1, dtype=torch.long, device=x.device)).view(
            1, self.inner_dim
        )  # [1, D]

        t0 = self.norm2(att0)
        t1_v = self.norm2(att1_v)
        t1_e = self.norm2(att1_e)
        att0 = att0 + self.mlp2(torch.cat((t0, pe2_s0.expand(t0.shape)), dim=-1))
        att1_v = att1_v + self.mlp2(
            torch.cat((t1_v, pe2_s1.expand(t1_v.shape)), dim=-1)
        )
        att1_e = att1_e + self.mlp2(
            torch.cat((t1_e, pe2_s1.expand(t1_e.shape)), dim=-1)
        )

        x_v = self.att0_dropout(att0) + self.att1_dropout(att1_v)
        x_e = self.att0_dropout(att0) + self.att1_dropout(att1_e)

        # MLP3
        pe3_l1 = self.pe3(torch.ones(1, dtype=torch.long, device=x.device)).view(
            1, self.inner_dim
        )
        if (
            self.max_l < len(edge_orders) and self.hyper_dropout == 0
        ) or self.force_broadcast:
            # do not use this when hypernetwork is stochastic
            indices = torch.arange(self.pe3.max_pos, device=x.device)
            pe3_l = self.pe3(indices)[edge_orders]  # [|E|, D]
        else:
            pe3_l = self.pe3(edge_orders)  # [|E|, D]

        t_v = self.norm3(x_v)
        t_e = self.norm3(x_e)
        x_v = x_v + self.mlp3(torch.cat((t_v, pe3_l1.expand(t_v.shape)), dim=-1))
        x_e = x_e + self.mlp3(torch.cat((t_e, pe3_l.expand(t_e.shape)), dim=-1))

        x = x_v, x_e
        x = self.b(x, edge_orders)
        return x

class SelfAttnE2V(nn.Module):
    def __init__(
        self,
        dim_in,
        dim_qk,
        n_heads,
        hypernet_info,
        inner_dim,
        pe_dim,
        hyper_dim,
        hyper_layers,
        hyper_dropout,
        force_broadcast,
        att0_dropout,
        att1_dropout,
    ):
        super().__init__()
        max_k, _, _ = hypernet_info
        self.max_k = max_k
        self.dim_in = dim_in
        self.dim_qk = dim_qk
        self.n_heads = n_heads
        self.dim_v = inner_dim
        self.inner_dim = inner_dim
        self.pe_dim = pe_dim
        self.hyper_dim = hyper_dim
        self.hyper_layers = hyper_layers
        self.hyper_dropout = hyper_dropout
        self.force_broadcast = force_broadcast
        self.dim_qk_head = dim_qk // n_heads if dim_qk >= n_heads else 1
        self.dim_v_head = self.dim_v // n_heads if self.dim_v >= n_heads else 1
        self.q = PositionalMLP(
            n_heads * self.dim_qk_head,
            2,
            pe_dim,
            hyper_dim,
            hyper_layers,
            hyper_dropout,
        )
        self.k = nn.Linear(self.dim_v, 2 * dim_qk)
        self.v = nn.Linear(self.dim_v, self.dim_v)

        self.mlp1 = MLP_inner(dim_in * 2, self.dim_v, hyper_dim, hyper_layers, hyper_dropout)
        self.mlp2 = MLP_inner(
            self.dim_v * 2, self.dim_v, hyper_dim, hyper_layers, hyper_dropout
        )
        self.mlp3 = MLP_inner(self.dim_v, dim_in, hyper_dim, hyper_layers, hyper_dropout)
        self.norm1 = nn.LayerNorm(inner_dim)
        self.norm2 = nn.LayerNorm(inner_dim)
        self.norm3 = nn.LayerNorm(inner_dim)
        # node-input only, don't need pe3
        self.pe1 = PositionalEncoding(dim_in, max_k + 1)
        self.pe2 = PositionalEncoding(inner_dim, 2)
        self.b = BiasV(dim_in)

        self.att0_dropout = nn.Dropout(att0_dropout)
        self.att1_dropout = nn.Dropout(att1_dropout)

    def reset_parameters(self):
        self.q.reset_parameters()
        self.k.reset_parameters()
        self.v.reset_parameters()
        self.mlp1.reset_parameters()
        self.mlp2.reset_parameters()
        self.mlp3.reset_parameters()
        self.norm1.reset_parameters()
        self.norm2.reset_parameters()
        self.norm3.reset_parameters()
        self.b.reset_parameters()

    def forward(self, x, ehnn_cache):
        """forward method
        :param x: [N, D]
        :param ehnn_cache:
        :return: [N, D], [|E|, D]
        """
        incidence = ehnn_cache["incidence"]
        edge_orders = ehnn_cache["edge_orders"]
        indices_with_nodes = ehnn_cache["indices_with_nodes"]

        # MLP1
        x_v, x_e = x
        pe1_k1 = self.pe1(torch.ones(1, dtype=torch.long, device=x_v.device)).view(
            1, self.dim_in
        )
        if (
            self.max_k < len(edge_orders) and self.hyper_dropout == 0
        ) or self.force_broadcast:
            # do not use this when hypernetwork is stochastic
            indices = torch.arange(self.pe1.max_pos, device=x_v.device)
            pe1_k = self.pe1(indices)[edge_orders]  # [|E|, D]
        else:
            pe1_k = self.pe1(edge_orders)  # [|E|, D]

        t_v = self.norm1(x_v)
        t_e = self.norm1(x_e)
        x_v = x_v + self.mlp1(torch.cat((t_v, pe1_k1.expand(t_v.shape)), dim=-1))
        x_e = x_e + self.mlp1(torch.cat((t_e, pe1_k.expand(t_e.shape)), dim=-1))

        q_0 = self.q(torch.zeros(1, dtype=torch.long, device=x_v.device)).view(
            self.n_heads, self.dim_qk_head
        )
        q_1 = self.q(torch.ones(1, dtype=torch.long, device=x_v.device)).view(
            self.n_heads, self.dim_qk_head
        )
        k_v, k_e = self.k(x_v), self.k(x_e)
        k_v_0, k_v_1 = k_v.split(self.dim_qk, -1)
        k_e_0, k_e_1 = k_e.split(self.dim_qk, -1)
        k_v_0 = torch.stack(k_v_0.split(self.dim_qk_head, -1), 1)  # [N, H, D/H]
        k_v_1 = torch.stack(k_v_1.split(self.dim_qk_head, -1), 1)  # [N, H, D/H]
        k_e_0 = torch.stack(k_e_0.split(self.dim_qk_head, -1), 1)  # [|E|, H, D/H]
        k_e_1 = torch.stack(k_e_1.split(self.dim_qk_head, -1), 1)  # [|E|, H, D/H]
        k_0 = torch.cat([k_e_0, k_v_0])
        v_v, v_e = self.v(x_v), self.v(x_e)
        v_v = torch.stack(v_v.split(self.dim_v_head, -1), 1)  # [N, H, Dv/H]
        v_e = torch.stack(v_e.split(self.dim_v_head, -1), 1)  # [|E|, H, Dv/H]

        # aggregation
        # either 0-overlap (global), or 1-overlap (local)
        v = torch.cat([v_e, v_v])  # [|E| + N, H, Dv/H]
        # 0-overlap
        # softmax (not using kernel)
        logit0 = torch.einsum("hd,ehd->eh", q_0, k_0) / math.sqrt(
            self.dim_qk_head
        )  # [|E| + N, H]
        alpha0 = torch.softmax(logit0, dim=0)  # [|E| + N, H]
        att0 = torch.einsum("eh,ehd->hd", alpha0, v)  # [H, Dv/H]
        att0 = torch.cat(att0.unbind(-2), -1)  # [Dv,]
        att0 = att0[None, :]  # [1, Dv]
        # 1-overlap
        logit1 = torch.einsum("hd,nhd->nh", q_1, torch.cat([k_e_1, k_v_1])) / math.sqrt(
            self.dim_qk_head
        )  # [|E| + N, H]
        n, e = incidence.size()
        logit1 = torch.sparse_coo_tensor(
            indices_with_nodes,
            logit1[indices_with_nodes[1]],
            size=(n, e + n, self.n_heads),
        )  # [N, |E| + N, H]
        alpha1 = torch.sparse.softmax(logit1, dim=1).coalesce()  # [N, |E| + N, H]
        att1 = util_sparse_einsum(alpha1, v)  # [N, H, Dv/H]
        att1 = torch.cat(att1.unbind(-2), -1)  # [N, Dv]

        # MLP2
        pe2_s0 = self.pe2(torch.zeros(1, dtype=torch.long, device=x_v.device)).view(
            1, self.inner_dim
        )  # [1, D]
        pe2_s1 = self.pe2(torch.ones(1, dtype=torch.long, device=x_v.device)).view(
            1, self.inner_dim
        )  # [1, D]

        t0 = self.norm2(att0)
        t1 = self.norm2(att1)
        att0 = att0 + self.mlp2(torch.cat((t0, pe2_s0.expand(t0.shape)), dim=-1))
        att1 = att1 + self.mlp2(torch.cat((t1, pe2_s1.expand(t1.shape)), dim=-1))

        # output has only 1-edges
        x = self.att0_dropout(att0) + self.att1_dropout(att1)
        # MLP3
        x = x + self.mlp3(self.norm3(x))  # [N, D']
        x = self.b(x)
        return x

class TransformerV2E(nn.Module):
    def __init__(
        self,
        dim_in,
        dim_qk,
        dim_ff,
        n_heads,
        dropout,
        hypernet_info,
        inner_dim,
        pe_dim,
        hyper_dim,
        hyper_layers,
        hyper_dropout,
        force_broadcast,
        att0_dropout,
        att1_dropout,
    ):
        super().__init__()
        self.ln = nn.LayerNorm(dim_in)
        self.attn = SelfAttnV2E(
            dim_in,
            dim_qk,
            n_heads,
            hypernet_info,
            inner_dim,
            pe_dim,
            hyper_dim,
            hyper_layers,
            hyper_dropout,
            force_broadcast,
            att0_dropout,
            att1_dropout,
        )
        self.ffn = nn.Sequential(
            nn.LayerNorm(dim_in),
            nn.Linear(dim_in, dim_ff),
            nn.Dropout(dropout, inplace=True),
            nn.ReLU(),
            nn.Linear(dim_ff, dim_in),
        )

    def reset_parameters(self):
        self.ln.reset_parameters()
        self.attn.reset_parameters()
        for layer in self.ffn:
            if isinstance(layer, nn.LayerNorm):
                layer.reset_parameters()
            if isinstance(layer, nn.Linear):
                layer.reset_parameters()

    def forward(self, x, ehnn_cache):
        h = self.ln(x)
        h_v, h_e = self.attn(h, ehnn_cache)
        x_v, x_e = (x + h_v, h_e)
        h_v, h_e = (self.ffn(x_v), self.ffn(x_e))
        x_v, x_e = (x_v + h_v, x_e + h_e)
        return x_v, x_e

class TransformerE2V(nn.Module):
    def __init__(
        self,
        dim_in,
        dim_qk,
        dim_ff,
        n_heads,
        dropout,
        hypernet_info,
        inner_dim,
        pe_dim,
        hyper_dim,
        hyper_layers,
        hyper_dropout,
        force_broadcast,
        att0_dropout,
        att1_dropout,
    ):
        super().__init__()
        self.ln = nn.LayerNorm(dim_in)
        self.attn = SelfAttnE2V(
            dim_in,
            dim_qk,
            n_heads,
            hypernet_info,
            inner_dim,
            pe_dim,
            hyper_dim,
            hyper_layers,
            hyper_dropout,
            force_broadcast,
            att0_dropout,
            att1_dropout,
        )
        self.ffn = nn.Sequential(
            nn.LayerNorm(dim_in),
            nn.Linear(dim_in, dim_ff),
            nn.Dropout(dropout, inplace=True),
            nn.ReLU(),
            nn.Linear(dim_ff, dim_in),
        )

    def reset_parameters(self):
        self.ln.reset_parameters()
        self.attn.reset_parameters()
        for layer in self.ffn:
            if isinstance(layer, nn.LayerNorm):
                layer.reset_parameters()
            if isinstance(layer, nn.Linear):
                layer.reset_parameters()

    def forward(self, x, ehnn_cache) -> torch.Tensor:
        x_v, x_e = x
        h = (self.ln(x_v), self.ln(x_e))
        h = self.attn(h, ehnn_cache)
        x = x_v + h
        h = self.ffn(x)
        x = x + h
        return x

class PositionalEncoding(nn.Module):
    """https://pytorch.org/tutorials/beginner/transformer_tutorial.html"""
    def __init__(self, dim, max_pos):
        super().__init__()
        self.max_pos = max_pos
        position = torch.arange(max_pos).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, dim, 2) * (-math.log(10000.0) / dim))
        pe = torch.zeros(max_pos, dim)
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        self.register_buffer('pe', pe)

    def forward(self, x: torch.LongTensor) -> torch.Tensor:
        return self.pe[x]

class PositionalMLP(nn.Module):
    def __init__(self, dim_out, max_pos, dim_pe, dim_hidden, n_layers, dropout):
        super().__init__()
        self.max_pos = max_pos
        self.pe = PositionalEncoding(dim_pe, max_pos)
        self.input = nn.Linear(dim_pe, dim_hidden)
        self.mlp = MLP_inner(dim_hidden, dim_out, dim_hidden, n_layers, dropout)

    def reset_parameters(self):
        self.input.reset_parameters()
        self.mlp.reset_parameters()

    def forward(self, x: torch.LongTensor) -> torch.Tensor:
        # LayerNorm: OK
        # Dropout: OK
        # Skip connections: not OK
        x = self.input(self.pe(x))
        x = self.mlp(x)
        return x

class PositionalWeight(nn.Module):
    def __init__(self, max_pos, dim_in, dim_out):
        super().__init__()
        self.max_pos = max_pos
        self.dim_in = dim_in
        self.dim_out = dim_out
        self.weights = nn.Parameter(torch.Tensor(max_pos + 1, dim_in, dim_out))

    def reset_parameters(self):
        stdv = 1. / math.sqrt(self.dim_out)
        self.weights.data.uniform_(-stdv, stdv)

    def forward(self, x: torch.LongTensor) -> torch.Tensor:
        return self.weights[x].view(x.size(0), -1)

class BiasE(nn.Module):
    def __init__(self, dim_out,
                 max_l, pe_dim, hyper_dim, hyper_layers, hyper_dropout, force_broadcast):
        super().__init__()
        self.dim_out = dim_out
        self.max_l = max_l
        self.pe_dim = pe_dim
        self.hyper_dim = hyper_dim
        self.hyper_layers = hyper_layers
        self.hyper_dropout = hyper_dropout
        self.force_broadcast = force_broadcast
        self.b = PositionalMLP(dim_out, max_l + 1, pe_dim, hyper_dim, hyper_layers, hyper_dropout)

    def reset_parameters(self):
        self.b.reset_parameters()

    def forward(self, x, edge_orders):
        x_v, x_e = x
        if (self.max_l and self.hyper_dropout == 0) or self.force_broadcast:
            # do not use this when hypernetwork is stochastic
            indices = torch.arange(self.b.max_pos, device=x_v.device)
            b = self.b(indices)[edge_orders]
        else:
            b = self.b(edge_orders)
        b_1 = self.b(torch.ones(1, dtype=torch.long, device=x_v.device)).view(1, self.dim_out)  # [D']
        return x_v + b_1, x_e + b

class BiasV(nn.Module):
    def __init__(self, dim_out):
        super().__init__()
        self.dim_out = dim_out
        self.b = nn.Parameter(torch.Tensor(1, dim_out))

    def reset_parameters(self):
        stdv = 1. / math.sqrt(self.dim_out)
        self.b.data.uniform_(-stdv, stdv)

    def forward(self, x):
        return x + self.b


import torch
import torch.nn as nn
from torch_scatter import scatter_add, scatter
from dhgbench.lib_models.HNN.utils import zeros,glorot,normalize_l2

# v1: X -> XW -> AXW -> norm
class UniGINConv(nn.Module):

    def __init__(self, args, in_channels, out_channels, heads=1, dropout=0., negative_slope=0.2):
        super().__init__()
        self.W = nn.Linear(in_channels, heads * out_channels, bias=False)
        self.heads = heads
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.negative_slope = negative_slope
        self.dropout = dropout
        self.eps = nn.Parameter(torch.Tensor([0.]))
        self.args = args 

    def reset_parameters(self):
        nn.init.xavier_uniform_(self.W.weight)
        zeros(self.eps)
    
    def forward(self, X, vertex, edges):
        N = X.shape[0]
        
        # v1: X -> XW -> AXW -> norm
        X = self.W(X) 

        Xve = X[vertex] # [nnz, C]
        Xe = scatter(Xve, edges, dim=0, reduce=self.args.first_aggregate) # [E, C]
        
        Xev = Xe[edges] # [nnz, C]
        Xv = scatter(Xev, vertex, dim=0, reduce='sum', dim_size=N) # [N, C]
        X = (1 + self.eps) * X + Xv 

        if self.args.use_norm:
            X = normalize_l2(X)

        return X,Xe

__all_convs__ = {
'UniGIN': UniGINConv,
}

class UniGNN(nn.Module):
    def __init__(self,num_features, num_targets, args):

        super(UniGNN, self).__init__()
        UniGNNConv = __all_convs__[args.method]
        self.num_layers = args.All_num_layers
        self.nhead=args.uni_heads
        self.hidden_dim = args.MLP_hidden
        
        self.conv_out = UniGNNConv(args, self.hidden_dim * self.nhead, num_targets, heads=1, dropout=args.attn_drop)
        self.convs = nn.ModuleList(
            [UniGNNConv(args, num_features, self.hidden_dim, heads=self.nhead, dropout=args.attn_drop)] +
            [UniGNNConv(args, self.hidden_dim * self.nhead, self.hidden_dim, heads=self.nhead, dropout=args.attn_drop) for _ in range(self.num_layers-2)]
        )
        act = {'relu': nn.ReLU(), 'prelu':nn.PReLU() }
        self.act = act[args.activation]
        self.input_drop = nn.Dropout(args.input_drop)
        self.dropout = nn.Dropout(args.dropout)

    def reset_parameters(self):
        for conv in self.convs:
            conv.reset_parameters()
    
    def forward(self, data):
        
        X=data.x
        V,E=data.hyperedge_index[0],data.hyperedge_index[1]
        
        X = self.input_drop(X)
        for conv in self.convs:
            X,_ = conv(X, V, E)
            X = self.act(X)
            X = self.dropout(X)

        X,Xe = self.conv_out(X, V, E)      
        return X,Xe


import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
from torch_scatter import scatter_add
from torch_geometric.nn.conv import MessagePassing
from torch_geometric.utils import softmax
from typing import Optional
from dhgbench.lib_models.HNN.utils import zeros,glorot

class HypergraphConv(MessagePassing):

    def __init__(self, in_channels, out_channels, symdegnorm=False, use_attention=False, heads=1,
                 concat=True, negative_slope=0.2, dropout=0, bias=True,
                 **kwargs):
        kwargs.setdefault('aggr', 'add')
        super(HypergraphConv, self).__init__(node_dim=0, **kwargs)

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.use_attention = use_attention
        self.symdegnorm = symdegnorm

        if self.use_attention:
            self.heads = heads
            self.concat = concat
            self.negative_slope = negative_slope
            self.dropout = dropout
            self.weight = nn.Parameter(
                torch.Tensor(in_channels, heads * out_channels))
            self.att = nn.Parameter(torch.Tensor(1, heads, 2 * out_channels))
        else:
            self.heads = 1
            self.concat = True
            self.weight = nn.Parameter(torch.Tensor(in_channels, out_channels))

        if bias and concat:
            self.bias = nn.Parameter(torch.Tensor(heads * out_channels))
        elif bias and not concat:
            self.bias = nn.Parameter(torch.Tensor(out_channels))
        else:
            self.register_parameter('bias', None)

        self.reset_parameters()

    def reset_parameters(self):
        glorot(self.weight)
        if self.use_attention:
            glorot(self.att)
        zeros(self.bias)

    def forward(self, x: Tensor, hyperedge_index: Tensor,
                hyperedge_weight: Optional[Tensor] = None) -> Tensor:
        r"""
        Args:
            x (Tensor): Node feature matrix :math:`\mathbf{X}`
            hyperedge_index (LongTensor): The hyperedge indices, *i.e.*
                the sparse incidence matrix
                :math:`\mathbf{H} \in {\{ 0, 1 \}}^{N \times M}` mapping from
                nodes to edges.
            hyperedge_weight (Tensor, optional): Sparse hyperedge weights
                :math:`\mathbf{W} \in \mathbb{R}^M`. (default: :obj:`None`)
        """
        num_nodes, num_edges = x.size(0), 0
        if hyperedge_index.numel() > 0:
            num_edges = int(hyperedge_index[1].max()) + 1

        if hyperedge_weight is None:
            hyperedge_weight = x.new_ones(num_edges,device=x.device)

        x = torch.matmul(x, self.weight)

        alpha = None
        if self.use_attention:
            assert num_edges <= num_edges
            x = x.view(-1, self.heads, self.out_channels)
            x_i, x_j = x[hyperedge_index[0]], x[hyperedge_index[1]]
            alpha = (torch.cat([x_i, x_j], dim=-1) * self.att).sum(dim=-1)
            alpha = F.leaky_relu(alpha, self.negative_slope)
            alpha = softmax(alpha, hyperedge_index[0], num_nodes=x.size(0))
            alpha = F.dropout(alpha, p=self.dropout, training=self.training)

        if not self.symdegnorm:
            D = scatter_add(hyperedge_weight[hyperedge_index[1]],
                            hyperedge_index[0], dim=0, dim_size=num_nodes)
            D = 1.0 / D
            D[D == float("inf")] = 0

            B = scatter_add(x.new_ones(hyperedge_index.size(1),device=x.device),
                            hyperedge_index[1], dim=0, dim_size=num_edges)  
            B = 1.0 / B
            B[B == float("inf")] = 0
            
            self.flow = 'source_to_target'
            edge_embed = self.propagate(hyperedge_index, x=x, norm=B, alpha=alpha,
                                 size=(num_nodes, num_edges))
            self.flow = 'target_to_source'
            out = self.propagate(hyperedge_index, x=edge_embed, norm=D, alpha=alpha,size=(num_nodes, num_edges))
            
        else:  # this correspond to HGNN
            D = scatter_add(hyperedge_weight[hyperedge_index[1]],
                            hyperedge_index[0], dim=0, dim_size=num_nodes)
            D = 1.0 / D**(0.5)
            D[D == float("inf")] = 0

            B = scatter_add(x.new_ones(hyperedge_index.size(1),device=x.device),
                            hyperedge_index[1], dim=0, dim_size=num_edges)
            B = 1.0 / B
            B[B == float("inf")] = 0

            x = D.unsqueeze(-1)*x
            self.flow = 'source_to_target'
            edge_embed = self.propagate(hyperedge_index, x=x, norm=B, alpha=alpha,
                                 size=(num_nodes, num_edges))

            self.flow = 'target_to_source'
            out = self.propagate(hyperedge_index,x=edge_embed, norm=D, alpha=alpha,size=(num_nodes, num_edges))

        if self.concat is True:
            out = out.view(-1, self.heads * self.out_channels)
            edge_embed = edge_embed.view(-1,self.heads * self.out_channels)
        else:
            out = out.mean(dim=1)
            edge_embed = edge_embed.mean(dim=-1)

        if self.bias is not None:
            out = out + self.bias

        return out,edge_embed 

    def message(self, x_j: Tensor, norm_i: Tensor, alpha: Tensor) -> Tensor:
        H, F = self.heads, self.out_channels
        
        out = norm_i.view(-1, 1, 1) * x_j.view(-1, H, F)

        if alpha is not None:
            out = alpha.view(-1, self.heads, 1) * out

        return out

    def __repr__(self):
        return "{}({}, {})".format(self.__class__.__name__, self.in_channels,
                                   self.out_channels)

class HCHA(nn.Module):

    def __init__(self, num_features, num_targets, args):
        super(HCHA, self).__init__()

        self.num_layers = args.All_num_layers
        self.dropout = args.dropout  # Note that default is 0.6
        self.symdegnorm = args.HCHA_symdegnorm
        self.hidden_dim = args.MLP_hidden

#       Note that add dropout to attention is default in the original paper
        self.convs = nn.ModuleList()
        if self.num_layers == 1:
            self.convs.append(HypergraphConv(num_features,
                               num_targets, self.symdegnorm))
        else:
            self.convs.append(HypergraphConv(num_features,
                                            self.hidden_dim, self.symdegnorm))
            for _ in range(self.num_layers-2):
                self.convs.append(HypergraphConv(
                    self.hidden_dim, self.hidden_dim, self.symdegnorm))
            # Output heads is set to 1 as default
            self.convs.append(HypergraphConv(
                self.hidden_dim, num_targets, self.symdegnorm))

    def reset_parameters(self):
        for conv in self.convs:
            conv.reset_parameters()

    def forward(self, data):

        # regular node classification

        x = data.x
        edge_index = data.hyperedge_index

        for i, conv in enumerate(self.convs[:-1]):
            x , e = conv(x,edge_index) 
            x = F.elu(x)
            x = F.dropout(x, p=self.dropout, training=self.training)

        x,e = self.convs[-1](x, edge_index)

        return x,e

    @torch.no_grad()
    def predict(self,data):
        self.eval()
        return self.forward(data)


import torch 
import torch.nn as nn
import torch.nn.functional as F
from dhgbench.lib_models.HNN.mlp import MLP

class PhenomNNMultiConv(nn.Module):

    def __init__(self, args):
        super(PhenomNNMultiConv, self).__init__()
        self.args = args
        self.lam0=args.lam0
        self.lam1=args.lam1
        self.alpha=args.alpha if args.alpha !=0 else 1/(1+args.lam0+args.lam1)
        self.num_steps=args.prop_step
        
    def forward(self, X, A, G):
        
        A_beta,A_gamma=A
        D_beta,D_gamma,I=G

        Y = Y0 = X

        Q_tild= self.lam0*D_beta+self.lam1*D_gamma + I

        # Iteratively compute new Y by equation (6) in the paper.
        for k in range(self.num_steps):
   
            Y_hat = self.lam0 * A_beta @ Y + Y0 + self.lam1 *  A_gamma @ Y 
            Y = F.relu((1 - self.alpha) * Y + self.alpha * (Q_tild ** -1) @ Y_hat)

        # we have linear out of this module
        return Y

class PhenomNN(nn.Module):
    def __init__(self, num_features, num_targets, args):
        super(PhenomNN, self).__init__()
        
        self.hidden_dim = args.MLP_hidden 
        self.norm_type=args.normalization
        self.dropout = args.dropout
        self.act_fn = nn.ReLU()
        
        self.projector = MLP(
            in_channels=num_features,
            hidden_channels=self.hidden_dim,
            out_channels=self.hidden_dim,
            num_layers=args.encoder_num_layers,
            dropout=self.dropout,
            Normalization=self.norm_type,
            InputNorm=False,
        )

        self.decoder = MLP(
            in_channels=self.hidden_dim,
            hidden_channels=args.decoder_hidden,
            out_channels=num_targets,
            num_layers=args.decoder_num_layers,
            dropout=self.dropout,
            Normalization=self.norm_type,
            InputNorm=False,
        )

        self.conv = PhenomNNMultiConv(args=args)

    def reset_parameters(self):

        self.projector.reset_parameters()
        self.decoder.reset_parameters()
        
    def forward(self, data):

        input,adj,G = data.x, data.adj, data.G

        x = F.dropout(input, self.dropout, training=self.training)
        x = F.relu(self.projector(x))
        x = self.conv(x, A=adj, G=G)
        x = F.dropout(x, self.dropout, training=self.training)
        x = self.decoder(x)
        return x,None


import math
import torch
import torch.nn as nn
from dhgbench.lib_models.HNN.mlp import MLP_inner

class EHNNLinearConv(nn.Module):
    def __init__(
        self,
        dim_in,
        dim_hidden,
        inner_dim,
        dropout,
        hypernet_info,
        pe_dim,
        hyper_dim,
        hyper_layers,
        hyper_dropout,
        force_broadcast,
        input_dropout,
        mlp_decoder,
    ):
        super().__init__()
        assert mlp_decoder
        self.input_dropout = nn.Dropout(input_dropout)
        self.lin_proj = nn.Linear(dim_in, dim_hidden)
        self.v2e_layer = LinearV2E(
            dim_hidden,
            dim_hidden,
            hypernet_info,
            inner_dim,
            pe_dim,
            hyper_dim,
            hyper_layers,
            hyper_dropout,
            force_broadcast,
        )
        self.e2v_layer = LinearE2V(
            dim_hidden,
            dim_hidden,
            hypernet_info,
            inner_dim,
            pe_dim,
            hyper_dim,
            hyper_layers,
            hyper_dropout,
            force_broadcast,
        )
        self.act = nn.ReLU(inplace=True)
        self.dropout = nn.Dropout(dropout)
        print(f"hypernet max input range: {hypernet_info}")

    def reset_parameters(self):
        self.lin_proj.reset_parameters()
        self.v2e_layer.reset_parameters()
        self.e2v_layer.reset_parameters()

    def forward(self, x, ehnn_cache) -> torch.Tensor:
        x = self.lin_proj(self.input_dropout(x))
        x_v, x_e = self.v2e_layer(x, ehnn_cache)
        x_v, x_e = self.act(x_v), self.act(x_e)
        x_v, x_e = self.dropout(x_v), self.dropout(x_e)
        x = (x_v, x_e)
        x = self.e2v_layer(x, ehnn_cache)
        return x, x_e


class LinearV2E(nn.Module):
    def __init__(self, dim_in, dim_out,
                 hypernet_info, inner_dim, pe_dim, hyper_dim, hyper_layers, hyper_dropout, force_broadcast):
        super().__init__()
        #assert dim_in == dim_out == inner_dim
        _, max_l, _ = hypernet_info
        self.dim_in = dim_in
        self.dim_out = dim_out
        self.inner_dim = inner_dim
        self.max_l = max_l
        self.pe_dim = pe_dim
        self.hyper_dim = hyper_dim
        self.hyper_layers = hyper_layers
        self.hyper_dropout = hyper_dropout
        self.force_broadcast = force_broadcast
        # node-input only, don't need pe1
        self.mlp1 = MLP_inner(dim_in, inner_dim, hyper_dim, hyper_layers, hyper_dropout)
        self.mlp2 = MLP_inner(inner_dim*2, inner_dim, hyper_dim, hyper_layers, hyper_dropout)
        self.mlp3 = MLP_inner(inner_dim*2, dim_out, hyper_dim, hyper_layers, hyper_dropout)
        self.norm1 = nn.LayerNorm(inner_dim)
        self.norm2 = nn.LayerNorm(inner_dim)
        self.norm3 = nn.LayerNorm(inner_dim)
        self.pe2 = PositionalEncoding(inner_dim, 2)
        self.pe3 = PositionalEncoding(inner_dim, max_l + 1)
        self.b = BiasE(dim_out, max_l, pe_dim, hyper_dim, hyper_layers, hyper_dropout, force_broadcast)

    def reset_parameters(self):
        self.mlp1.reset_parameters()
        self.mlp2.reset_parameters()
        self.mlp3.reset_parameters()
        self.norm1.reset_parameters()
        self.norm2.reset_parameters()
        self.norm3.reset_parameters()
        self.b.reset_parameters()

    def forward(self, x, ehnn_cache):
        incidence = ehnn_cache['incidence']
        edge_orders = ehnn_cache['edge_orders']
        prefix_normalizer = ehnn_cache['prefix_normalizer']
        # MLP1
        x = x + self.mlp1(self.norm1(x))  # [N, D]
        # aggregation
        x0 = x.mean(dim=0, keepdim=True)  # [1, D]
        x1_v = x  # [N, D]
        x1_e = (incidence.t() @ x / prefix_normalizer[:, None])  # [|E|, D]
        # MLP2
        pe2_s0 = self.pe2(torch.zeros(1, dtype=torch.long, device=x.device)).view(1, self.inner_dim)  # [1, D]
        pe2_s1 = self.pe2(torch.ones(1, dtype=torch.long, device=x.device)).view(1, self.inner_dim)  # [1, D]

        t0 = self.norm2(x0)
        x0 = x0 + self.mlp2(torch.cat((t0 , pe2_s0), dim=-1))

        t1_v = self.norm2(x1_v)
        t1_e = self.norm2(x1_e)
        x1_v = x1_v + self.mlp2(torch.cat((t1_v , pe2_s1.expand(t1_v.shape)), dim=-1))
        x1_e = x1_e + self.mlp2(torch.cat((t1_e , pe2_s1.expand(t1_e.shape)), dim=-1))

        x_v = x0 + x1_v
        x_e = x0 + x1_e
        # MLP3
        pe3_l1 = self.pe3(torch.ones(1, dtype=torch.long, device=x.device)).view(1, self.inner_dim)
        if (self.max_l < len(edge_orders) and self.hyper_dropout == 0) or self.force_broadcast:
            # do not use this when hypernetwork is stochastic
            indices = torch.arange(self.pe3.max_pos, device=x.device)
            pe3_l = self.pe3(indices)[edge_orders]  # [|E|, D]
        else:
            pe3_l = self.pe3(edge_orders)  # [|E|, D]

        t_v = self.norm3(x_v)
        t_e = self.norm3(x_e)
        x_v = x_v + self.mlp3(torch.cat((t_v , pe3_l1.expand(t_v.shape)), dim=-1))
        x_e = x_e + self.mlp3(torch.cat((t_e , pe3_l.expand(t_e.shape)), dim=-1))

        x = x_v, x_e
        x = self.b(x, edge_orders)
        return x 

class LinearE2V(nn.Module):
    def __init__(self, dim_in, dim_out,
                 hypernet_info, inner_dim, pe_dim, hyper_dim, hyper_layers, hyper_dropout, force_broadcast):
        super().__init__()
        # assert dim_in == dim_out == inner_dim
        max_k, _, _ = hypernet_info
        self.dim_in = dim_in
        self.dim_out = dim_out
        self.inner_dim = inner_dim
        self.max_k = max_k
        self.pe_dim = pe_dim
        self.hyper_dim = hyper_dim
        self.hyper_layers = hyper_layers
        self.hyper_dropout = hyper_dropout
        self.force_broadcast = force_broadcast
        self.mlp1 = MLP_inner(dim_in*2, inner_dim, hyper_dim, hyper_layers, hyper_dropout)
        self.mlp2 = MLP_inner(inner_dim*2, inner_dim, hyper_dim, hyper_layers, hyper_dropout)
        self.mlp3 = MLP_inner(inner_dim, dim_out, hyper_dim, hyper_layers, hyper_dropout)
        self.norm1 = nn.LayerNorm(inner_dim)
        self.norm2 = nn.LayerNorm(inner_dim)
        self.norm3 = nn.LayerNorm(inner_dim)
        # node-input only, don't need pe3
        self.pe1 = PositionalEncoding(dim_in, max_k + 1)
        self.pe2 = PositionalEncoding(inner_dim, 2)
        self.b = BiasV(dim_out)

    def reset_parameters(self):
        self.mlp1.reset_parameters()
        self.mlp2.reset_parameters()
        self.mlp3.reset_parameters()
        self.norm1.reset_parameters()
        self.norm2.reset_parameters()
        self.norm3.reset_parameters()
        self.b.reset_parameters()

    def forward(self, x, ehnn_cache) -> torch.Tensor:
        incidence = ehnn_cache['incidence']
        edge_orders = ehnn_cache['edge_orders']
        suffix_normalizer = ehnn_cache['suffix_normalizer']
        x_v, x_e = x
        # MLP1
        pe1_k1 = self.pe1(torch.ones(1, dtype=torch.long, device=x_v.device)).view(1, self.dim_in)
        if (self.max_k < len(edge_orders) and self.hyper_dropout == 0) or self.force_broadcast:
            # do not use this when hypernetwork is stochastic
            indices = torch.arange(self.pe1.max_pos, device=x_v.device)
            pe1_k = self.pe1(indices)[edge_orders]  # [|E|, D]
        else:
            pe1_k = self.pe1(edge_orders)  # [|E|, D]

        t_v = self.norm1(x_v)
        t_e = self.norm1(x_e)
        x_v = x_v + self.mlp1(torch.cat((t_v, pe1_k1.expand(t_v.shape)), dim=-1))
        x_e = x_e + self.mlp1(torch.cat((t_e , pe1_k.expand(t_e.shape)), dim=-1))

        # aggregation
        x0 = torch.cat((x_v, x_e)).mean(dim=0, keepdim=True)  # [1, D]
        x1 = ((x_v + incidence @ x_e) / (1 + suffix_normalizer[:, None]))  # [N, D]
        # MLP2
        pe2_s0 = self.pe2(torch.zeros(1, dtype=torch.long, device=x_v.device)).view(1, self.inner_dim)  # [1, D]
        pe2_s1 = self.pe2(torch.ones(1, dtype=torch.long, device=x_v.device)).view(1, self.inner_dim)  # [1, D]

        t0 = self.norm2(x0)
        t1 = self.norm2(x1)
        x0 = x0 + self.mlp2(torch.cat((t0 , pe2_s0.expand(t0.shape)), dim=-1))
        x1 = x1 + self.mlp2(torch.cat((t1 , pe2_s1.expand(t1.shape)), dim=-1))

        x = x0 + x1
        # MLP3
        x = x + self.mlp3(self.norm3(x))  # [N, D']
        x = self.b(x)
        return x

class PositionalEncoding(nn.Module):
    """https://pytorch.org/tutorials/beginner/transformer_tutorial.html"""
    def __init__(self, dim, max_pos):
        super().__init__()
        self.max_pos = max_pos
        position = torch.arange(max_pos).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, dim, 2) * (-math.log(10000.0) / dim))
        pe = torch.zeros(max_pos, dim)
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        self.register_buffer('pe', pe)

    def forward(self, x: torch.LongTensor) -> torch.Tensor:
        return self.pe[x]

class PositionalMLP(nn.Module):
    def __init__(self, dim_out, max_pos, dim_pe, dim_hidden, n_layers, dropout):
        super().__init__()
        self.max_pos = max_pos
        self.pe = PositionalEncoding(dim_pe, max_pos)
        self.input = nn.Linear(dim_pe, dim_hidden)
        self.mlp = MLP_inner(dim_hidden, dim_out, dim_hidden, n_layers, dropout)

    def reset_parameters(self):
        self.input.reset_parameters()
        self.mlp.reset_parameters()

    def forward(self, x: torch.LongTensor) -> torch.Tensor:
        # LayerNorm: OK
        # Dropout: OK
        # Skip connections: not OK
        x = self.input(self.pe(x))
        x = self.mlp(x)
        return x

class PositionalWeight(nn.Module):
    def __init__(self, max_pos, dim_in, dim_out):
        super().__init__()
        self.max_pos = max_pos
        self.dim_in = dim_in
        self.dim_out = dim_out
        self.weights = nn.Parameter(torch.Tensor(max_pos + 1, dim_in, dim_out))

    def reset_parameters(self):
        stdv = 1. / math.sqrt(self.dim_out)
        self.weights.data.uniform_(-stdv, stdv)

    def forward(self, x: torch.LongTensor) -> torch.Tensor:
        return self.weights[x].view(x.size(0), -1)

class BiasE(nn.Module):
    def __init__(self, dim_out,
                 max_l, pe_dim, hyper_dim, hyper_layers, hyper_dropout, force_broadcast):
        super().__init__()
        self.dim_out = dim_out
        self.max_l = max_l
        self.pe_dim = pe_dim
        self.hyper_dim = hyper_dim
        self.hyper_layers = hyper_layers
        self.hyper_dropout = hyper_dropout
        self.force_broadcast = force_broadcast
        self.b = PositionalMLP(dim_out, max_l + 1, pe_dim, hyper_dim, hyper_layers, hyper_dropout)

    def reset_parameters(self):
        self.b.reset_parameters()

    def forward(self, x, edge_orders):
        x_v, x_e = x
        if (self.max_l and self.hyper_dropout == 0) or self.force_broadcast:
            # do not use this when hypernetwork is stochastic
            indices = torch.arange(self.b.max_pos, device=x_v.device)
            b = self.b(indices)[edge_orders]
        else:
            b = self.b(edge_orders)
        b_1 = self.b(torch.ones(1, dtype=torch.long, device=x_v.device)).view(1, self.dim_out)  # [D']
        return x_v + b_1, x_e + b

class BiasV(nn.Module):
    def __init__(self, dim_out):
        super().__init__()
        self.dim_out = dim_out
        self.b = nn.Parameter(torch.Tensor(1, dim_out))

    def reset_parameters(self):
        stdv = 1. / math.sqrt(self.dim_out)
        self.b.data.uniform_(-stdv, stdv)

    def forward(self, x):
        return x + self.b


import argparse
import os
import yaml
import dhgbench
from dhgbench.lib_dataset import _single_datasets_,_multi_datasets_

def update_from_dict(obj, updates):
    for key, value in updates.items():
        # set higher priority from command line as we explore some factors
        if key in ['init'] and obj.init is not None:
            continue
        setattr(obj, key, value)

# recommend hyperparameters here
# def method_config(args):

#     if args.is_default:
#         config_name = 'default'
#     else:
#         config_name = args.dname
#     try:
#         # conf_dt = json.load(open(f"{os.path.join('./', 'lib_configs', args.method.lower(), config_name)}.json")) 
#         task_prefix=args.task_type.split('_')[0]+'_yamls'
#         conf_dt = yaml.safe_load(open(f"{os.path.join('./', 'lib_yamls', task_prefix,'config_'+args.method.lower())}.yaml"))[config_name] 
#         update_from_dict(args, conf_dt)
#     except:
#         print('No config file found or error in json format, please use method_config(args)')

#     return args
def method_config(args):

    if args.is_default:
        config_name = 'default'
    else:
        config_name = args.dname

    # -------- ① 找到包根目录 -----------
    pkg_root = os.path.dirname(dhgbench.__file__)  
    # 例如 /data1/.../site-packages/dhgbench

    # -------- ② 推导 YAML 所在目录 -----------
    task_prefix = args.task_type.split('_')[0] + '_yamls'  
    # node_cls → node_yamls
    # edge_pred → edge_yamls
    # hg_cls → hg_yamls

    yaml_path = os.path.join(
        pkg_root, "lib_yamls", task_prefix, f"config_{args.method.lower()}.yaml"
    )

    # -------- ③ 安全读取 -----------
    try:
        with open(yaml_path, "r") as f:
            conf_dt = yaml.safe_load(f)[config_name]
        update_from_dict(args, conf_dt)
        print(f"[DHGBench] Loaded config: {yaml_path}")
    except Exception as e:
        print(f"[DHGBench] Failed to load config: {yaml_path}")
        print(e)
        print("Please check method_config(args)")
    
    return args

def str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')

def set_task_args(args):
    
    if args.task_type == 'node_cls':
        if args.dname not in _single_datasets_:
            raise ValueError('The dataset is not suitable for node classification')
        args.add_self_loop=True 
        args.train_prop,args.valid_prop = 0.5,0.25
        args.early_stop = False
    elif args.task_type == 'hg_cls':
        if args.dname not in _multi_datasets_:
            raise ValueError('The datasets is not suitable for hypergraph classification')
        args.add_self_loop=False
        args.train_prop,args.valid_prop = 0.8,0.1
        args.early_stop = True
        if args.method in ['EHNN','TMPHN']:
            raise ValueError(f'{args.method} is not supoorted for hypergraph classification task') 
    else:
        if args.dname not in _single_datasets_:
            raise ValueError('The dataset is not suitable for edge prediction')

        if args.method in ['HyperND']:
            args.add_self_loop=True 
        elif args.method in ['DPHGNN','LEGCN','PhenomNN','HJRL','TFHNN','HNHN','AllSetformer'] and args.dname in ['pokec']:
            args.add_self_loop=True
        elif args.method in ['TMPHN']:
            if args.dname in ['pokec']:
                args.add_self_loop=True
            else:
                args.add_self_loop=False
            args.device='cpu' 
        else:
            args.add_self_loop=False
        
        args.train_prop,args.valid_prop = 0.6,0.2
        args.early_stop = True
    
    return args

def parameter_parser():
    """
    A method to parse up command line parameters.
    The default hyper-parameters give a good quality representation without grid search.
    """
    parser = argparse.ArgumentParser()

    ######################### general parameters ################################
    '''
    Semi-supervised setting: Train/Valid/Test: 50/25/25
    
    '''
    parser.add_argument('--train_prop', type=float, default=0.5)
    parser.add_argument('--valid_prop', type=float, default=0.25)
    parser.add_argument('--dname', default='pubmed',choices=['cora','citeseer','pubmed',
                                                            'coauthor_cora','coauthor_dblp',
                                                            '20newsW100', 'ModelNet40', 'zoo','NTU2012', 'Mushroom',
                                                            'yelp','walmart-trips-100','house-committees-100',
                                                            'actor','amazon','pokec','twitch',
                                                            'german','bail','credit',
                                                            'amazon_review','magpm','trivago','ogbn_mag',
                                                            "RHG_3", "RHG_10", "RHG_table", "RHG_pyramid",
                                                            "IMDB_dir_form", "IMDB_dir_genre",
                                                            "IMDB_wri_form", "IMDB_wri_genre",
                                                            "stream_player","twitter_friend"])
    
    parser.add_argument('--task_type',default='hg_cls',choices=['node_cls','edge_pred','hg_cls'])
    parser.add_argument('--is_default',default=False)
    parser.add_argument('--use_processed', default=True)
    parser.add_argument('--method', default='MLP') 
    
    parser.add_argument('--device', default='cuda:0')
    parser.add_argument('--num_seeds', type=int, default=5)
    parser.add_argument('--epochs', default=5, type=int)
    parser.add_argument('--dropout', default=0.5, type=float)
    parser.add_argument('--lr', default=0.001, type=float) # []
    parser.add_argument('--wd', default=0.0, type=float)
    parser.add_argument('--clip_grad',default=False,type=bool)
    parser.add_argument('--clip_thresh',default=5.0,type=float)
    parser.add_argument('--num_splits',type=int,default=10)
    parser.add_argument('--mem_verbose',default=True)
    parser.add_argument('--mem_display_step',default=100)
    parser.add_argument('--display_step', type=int, default=10)
    parser.add_argument('--eval_verbose',default=True)
    
    parser.add_argument('--embedding_mode',default=False,type=bool) 
    parser.add_argument('--embedding_hidden',default=128,type=int) 
    
    parser.add_argument('--normtype', default='all_one') # ['all_one','deg_half_sym']
    parser.add_argument('--add_self_loop', action='store_false')
    parser.add_argument('--exclude_self', action='store_true')
    
    parser.add_argument('--edge_save_dir', action='store_true',default=f'./lib_edge_splits/') 
    parser.add_argument('--edge_batch_size', action='store_true',default=512)
    parser.add_argument('--e_embed_hidden',default=64)
    parser.add_argument('--e_embed_layer',default=2) 
    parser.add_argument('--e_embed_dropout',default=0.2) 
    parser.add_argument('--e_embed_norm',default='ln')
    parser.add_argument('--aggr_mode',default='max',choices=['max','mean','maxmin'])
    parser.add_argument('--ns_method',default='mixed',choices=['mns','sns','cns','mixed']) 
    parser.add_argument('--edge_aggr',default='group',choices=['group','single']) 
    
    parser.add_argument('--hg_batch_size',default=256) 
    parser.add_argument('--pooling',default='mean')
    parser.add_argument('--g_embed_hidden',default=128) 
    parser.add_argument('--g_embed_layer',default=2) 
    parser.add_argument('--g_embed_dropout',default=0.2) 
    parser.add_argument('--g_embed_norm',default='ln') 
    parser.add_argument('--use_weighted_loss',default=False)
    parser.add_argument('--early_stop',default=True) 

    parser.add_argument('--is_perturbed',default=True) 
    parser.add_argument('--is_poison',default=True) 
    parser.add_argument('--pert_mode',default='drop_incidence',choices=['spar_feat','noise_feat',
                                                                    'drop_incidence','add_incidence',
                                                                    'spar_label','flip_label'])
    parser.add_argument('--pert_p',default=0.25)

    parser.add_argument('--feature_noise', default='0.6', type=str)
    
    parser.add_argument("--data_dir", type=str, default='./data',  help="Path to root directory of datasets")

    
    parser.set_defaults(add_self_loop=False)
    parser.set_defaults(exclude_self=False)

    args = parser.parse_args()

    return args




import random
import numpy as np
import torch
from torch import Tensor
from torch_geometric.typing import OptTensor
from torch_scatter import scatter_add
from torch_geometric.utils import dropout_adj

def mask_to_index(index, size):
    all_idx = np.arange(size)
    return all_idx[index]

def mean_std_metrics(metrics):
    
    metrics = np.array(metrics)
    metrics_dim = metrics.shape[-1]
    metrics = metrics.reshape(-1, metrics_dim) 
    
    metrics_mean = list(np.mean(metrics, axis=0))
    metrics_std = list(np.std(metrics, axis=0))
    
    return metrics_mean, metrics_std

def result_printer(metrics,name):
    
    metrics_dim = np.array(metrics).shape[-1]
    metrics_mean,metrics_std = mean_std_metrics(metrics)
    
    if metrics_dim == 1:
        print(f'{name}: {metrics_mean[0]:.4f}+-{metrics_std[0]:.2f}')
    else:
        print(f'train_{name}: {metrics_mean[0]:.4f}+-{metrics_std[0]:.2f}, val_{name}: {metrics_mean[1]:.4f}+-{metrics_std[1]:.2f}, test_{name}: {metrics_mean[2]:.4f}+-{metrics_std[2]:.2f}')

def fix_seed(seed):
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)
    random.seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = True

# ------------ hypergraph add_self_loop ----------------------------------

def relabel_hyperedge_index(hyperedge_index):

    node_idx = hyperedge_index[0]
    hyperedge_idx = hyperedge_index[1]

    unique_ids, new_ids = torch.unique(hyperedge_idx, return_inverse=True)

    new_hyperedge_index = torch.stack([node_idx, new_ids], dim=0)

    return new_hyperedge_index, unique_ids

def add_self_loop_hyperedges(hyperedge_index: torch.Tensor, num_nodes: int):

    device = hyperedge_index.device

    num_existing_edges = hyperedge_index[1].max().item() + 1  

    self_loop_node_indices = torch.arange(num_nodes, device=device)
    self_loop_edge_indices = torch.arange(num_existing_edges, num_existing_edges + num_nodes, device=device)

    self_loop_index = torch.stack([self_loop_node_indices, self_loop_edge_indices], dim=0)  

    new_hyperedge_index = torch.cat([hyperedge_index, self_loop_index], dim=1)

    return new_hyperedge_index

# ------------ hypergraph/graph augmentation methods ----------------------

def drop_features(x: Tensor, p: float):
    drop_mask = torch.empty((x.size(1), ), dtype=torch.float32, device=x.device).uniform_(0, 1) < p
    x = x.clone()
    x[:, drop_mask] = 0
    return x

def filter_incidence(row: Tensor, col: Tensor, hyperedge_attr: OptTensor, mask: Tensor):
    return row[mask], col[mask], None if hyperedge_attr is None else hyperedge_attr[mask]

def drop_incidence(hyperedge_index: Tensor, p: float = 0.2):
    if p == 0.0:
        return hyperedge_index
    
    row, col = hyperedge_index
    mask = torch.rand(row.size(0), device=hyperedge_index.device) >= p
    
    row, col, _ = filter_incidence(row, col, None, mask)
    hyperedge_index = torch.stack([row, col], dim=0)
    return hyperedge_index

def drop_nodes(hyperedge_index: Tensor, num_nodes: int, num_edges: int, p: float):
    if p == 0.0:
        return hyperedge_index

    drop_mask = torch.rand(num_nodes, device=hyperedge_index.device) < p
    drop_idx = drop_mask.nonzero(as_tuple=True)[0]

    H = torch.sparse_coo_tensor(hyperedge_index, \
        hyperedge_index.new_ones((hyperedge_index.shape[1],)), (num_nodes, num_edges)).to_dense()
    H[drop_idx, :] = 0
    hyperedge_index = H.to_sparse().indices()

    return hyperedge_index

def drop_hyperedges(hyperedge_index: Tensor, num_nodes: int, num_edges: int, p: float):
    if p == 0.0:
        return hyperedge_index

    drop_mask = torch.rand(num_edges, device=hyperedge_index.device) < p
    drop_idx = drop_mask.nonzero(as_tuple=True)[0]

    H = torch.sparse_coo_tensor(hyperedge_index, \
        hyperedge_index.new_ones((hyperedge_index.shape[1],)), (num_nodes, num_edges)).to_dense()
    H[:, drop_idx] = 0
    hyperedge_index = H.to_sparse().indices()

    return hyperedge_index

def valid_node_edge_mask(hyperedge_index: Tensor, num_nodes: int, num_edges: int):
    ones = hyperedge_index.new_ones(hyperedge_index.shape[1])
    Dn = scatter_add(ones, hyperedge_index[0], dim=0, dim_size=num_nodes)
    De = scatter_add(ones, hyperedge_index[1], dim=0, dim_size=num_edges)
    node_mask = Dn != 0
    edge_mask = De != 0
    return node_mask, edge_mask

def common_node_edge_mask(hyperedge_indexs: list[Tensor], num_nodes: int, num_edges: int):
    hyperedge_weight = hyperedge_indexs[0].new_ones(num_edges)
    node_mask = hyperedge_indexs[0].new_ones((num_nodes,)).to(torch.bool)
    edge_mask = hyperedge_indexs[0].new_ones((num_edges,)).to(torch.bool)

    for index in hyperedge_indexs:
        Dn = scatter_add(hyperedge_weight[index[1]], index[0], dim=0, dim_size=num_nodes)
        De = scatter_add(index.new_ones(index.shape[1]), index[1], dim=0, dim_size=num_edges)
        node_mask &= Dn != 0
        edge_mask &= De != 0
    return node_mask, edge_mask

def hyperedge_index_masking(hyperedge_index, num_nodes, num_edges, node_mask, edge_mask):
    if node_mask is None and edge_mask is None:
        return hyperedge_index

    H = torch.sparse_coo_tensor(hyperedge_index, \
        hyperedge_index.new_ones((hyperedge_index.shape[1],)), (num_nodes, num_edges)).to_dense()
    if node_mask is not None and edge_mask is not None:
        masked_hyperedge_index = H[node_mask][:, edge_mask].to_sparse().indices()
    elif node_mask is None and edge_mask is not None:
        masked_hyperedge_index = H[:, edge_mask].to_sparse().indices()
    elif node_mask is not None and edge_mask is None:
        masked_hyperedge_index = H[node_mask].to_sparse().indices()
    return masked_hyperedge_index




import os
import copy
import numpy as np
#from parameter_parser import parameter_parser
import torch
from collections import defaultdict
from dhgbench.lib_utils.utils import fix_seed,result_printer,mean_std_metrics
from dhgbench.lib_utils.train_agent import Trainer
from dhgbench.lib_utils.eval_agent import Evaluator
from dhgbench.lib_models.HNN import HCHA,HyperGCN,HNHN,SetGNN,UniGNN,UniGCNII,LEGCN,HyperND,EquivSetGNN,\
                            PlainUnigencoder,HJRL,SheafHyperGNN,EHNN,TMPHN,PhenomNN,PhenomNNS,DPHGNN,TFHNN,PlainMLP,HyperGT,CEGCN,CEGAT

from dhgbench.lib_dataset.data_perturbation import perturbation
from dhgbench.lib_dataset.edge_loaders import generate_edge_loaders,generate_split_hyperedges
from dhgbench.lib_dataset.hg_loaders import generate_split_hypergraphs,generate_hg_loaders
from dhgbench.lib_utils.aggregator import EdgePredictor,MeanAggregator,MaxminAggregator,MaxAggregator,HyperGPredictor
from dhgbench.lib_utils.metrics import aggr_metrics,avg_result_printer_edge

class ExpAgent:
    
    def __init__(self,args,**kwargs):
        """
        Overall pipline for different kinds of models
        """
        self.args = args
        self.device=args.device
        self.trainer=Trainer(args)
        self.evaluator=Evaluator(args)
        self.train_times = []
    
    def edge_pred_train_eval(self,data):
        
        metrics_dict = {'train':defaultdict(list),'val':defaultdict(list),'test':defaultdict(list)}
        
        for seed in range(self.args.num_seeds):
            
            fix_seed(seed) 
            
            file_path = f"{self.args.edge_save_dir}{self.args.dname}/split_{seed}.pt"
            
            if not os.path.exists(file_path):
                generate_split_hyperedges(data,self.args,seed)
                
            data_dict = torch.load(file_path, weights_only=False)
            
            batch_loaders = generate_edge_loaders(data_dict,self.args)
            
            self.args.embedding_mode = True 
            encoder = parse_model(self.args,data) 
            
            if self.args.aggr_mode=='maxmin':
                aggregator = MaxminAggregator(self.args) 
            elif self.args.aggr_mode=='mean':
                aggregator = MeanAggregator(self.args) 
            elif self.args.aggr_mode=='max':
                aggregator = MaxAggregator(self.args) 
            
            model = EdgePredictor(encoder,aggregator,self.args)
            if self.args.method == 'TMPHN':
                model.aggregator = model.aggregator.to(self.args.device)
            else:
                model = model.to(self.args.device)
            

            model = self.trainer.training(model,data,self.args,seed_split=batch_loaders,task_type='edge_pred')

            if self.args.eval_verbose:
                print(f'------------------------------[Seed {seed}]-----------------------------------')
                result=self.evaluator.evaluate(model,data,seed_split=batch_loaders,task_type='edge_pred',verbose=True)
                metrics_dict = aggr_metrics(metrics_dict,result) 
                print(f'------------------------------------------------------------------------------')
            else:
                result=self.evaluator.evaluate(model,data,seed_split=batch_loaders,task_type='edge_pred',verbose=True)
                metrics_dict = aggr_metrics(metrics_dict,result) 

        print(f'---------------------------------[Final]--------------------------------------')
        avg_result_printer_edge(metrics_dict)
        print(f'------------------------------------------------------------------------------')

    def node_cls_train_eval(self,data):
        
        metrics_dict=defaultdict(list)

        # 2. 多轮随机数种子实验
        for seed in range(self.args.num_seeds):
            
            # 1. 固定随机数种子
            fix_seed(seed) 
            
            # 2. 节点分类数据集随机划分
            masks=data.generate_random_split(train_ratio=self.args.train_prop,val_ratio=self.args.valid_prop,seed=seed)

            # 2.1 标签鲁棒性测试
            if self.args.is_perturbed:
                if self.args.pert_mode in ['spar_label','flip_label']:
                    if self.args.pert_mode == 'spar_label':
                        masks = perturbation(data,mode=self.args.pert_mode,p=self.args.pert_p,masks=masks)
                    elif self.args.pert_mode == 'flip_label':
                        data = perturbation(data,mode=self.args.pert_mode,p=self.args.pert_p,masks=masks)
                    else:
                        raise ValueError('Unimplemented perturbation mode for label robustness')

            # 3. 初始化模型
            model = parse_model(self.args,data)
            if self.args.method == 'TMPHN':
                pass
            else:
                model = model.to(self.args.device)

            # 4. 模型训练[调用trainer类]
            self.trainer.training(model,data,self.args,seed_split=masks,task_type='node_cls')
            
            self.train_times.append(self.trainer.train_time)

            # Evasion Attack
            if self.args.is_perturbed and not self.args.is_poison:
                test_data = data.evasion_data
            else:
                test_data = data

            if self.args.eval_verbose:
                print(f'------------------------------[Seed {seed}]-----------------------------------')
                # 5. 单一轮次模型评估[调用evaluator类]
                result=self.evaluator.evaluate(model,test_data,seed_split=masks,task_type='node_cls',verbose=True)
                print(f'------------------------------------------------------------------------------')
            else:
                result=self.evaluator.evaluate(model,test_data,seed_split=masks,task_type='node_cls',verbose=False)
            
            for m in result:
                metrics_dict[m].append(result[m])
            
        print(f'---------------------------------[Final]--------------------------------------')
        # 多轮随机数取均值和标准差
        self.test_dict = defaultdict(list) # 记录每个指标在测试集上的效果
        for m in metrics_dict:
            result_printer(metrics_dict[m],m)
            metrics_mean, metrics_std = mean_std_metrics(metrics_dict[m])
            self.test_dict[m].extend([metrics_mean[-1],metrics_std[-1]])
        print(f'Avg Training Time: {np.mean(self.train_times):2f}')
        print(f'------------------------------------------------------------------------------')

    def hg_cls_train_eval(self,data):
        
        metrics_dict=defaultdict(list)
        
        for seed in range(self.args.num_seeds):
            
            fix_seed(seed) 
            
            train_set,val_set,test_set = generate_split_hypergraphs(data,self.args.train_prop,self.args.valid_prop,seed)
            batch_loaders = generate_hg_loaders(train_set,val_set,test_set,self.args)
            
            self.args.embedding_mode = True 
            encoder = parse_model(self.args,data)

            model = HyperGPredictor(encoder,data.num_classes,self.args)
            if self.args.method == 'TMPHN':
                model.classifer = model.aggregator.to(self.args.device)
            else:
                model = model.to(self.args.device)
                     
            model = self.trainer.training(model,data,self.args,seed_split=batch_loaders,task_type='hg_cls')
            
            if self.args.eval_verbose:
                print(f'------------------------------[Seed {seed}]-----------------------------------')
                result=self.evaluator.evaluate(model,data,seed_split=batch_loaders,task_type='hg_cls',verbose=True)
                print(f'------------------------------------------------------------------------------')
            else:
                result=self.evaluator.evaluate(model,data,seed_split=batch_loaders,task_type='hg_cls',verbose=False)

            for m in result:
                metrics_dict[m].append(result[m])

        print(f'---------------------------------[Final]--------------------------------------')
        for m in metrics_dict:
            result_printer(metrics_dict[m],m)
        print(f'------------------------------------------------------------------------------')
        
    def running(self,task_type,data):
        
        if task_type == 'node_cls':
            self.node_cls_train_eval(data)
        elif task_type == 'edge_pred':
            self.edge_pred_train_eval(data)
        elif task_type == 'hg_cls':
            self.hg_cls_train_eval(data)
        else:
            raise NotImplementedError

def parse_model(args, data):
    
    if args.embedding_mode:
        num_targets=args.embedding_hidden
    else:
        num_targets=data.num_classes
    
    # --------- Hypergraph Semi-supervised Models --------------------
    
    if args.method == 'AllSetformer':
        if args.LearnMask:
            model = SetGNN(data.num_features, num_targets, args, data.norm)
        else:
            model = SetGNN(data.num_features, num_targets, args)
    elif args.method == 'AllDeepSets':
        args.PMA = False 
        args.aggregate = 'add'
        if args.LearnMask:
            model = SetGNN(data.num_features, num_targets, args, data.norm)
        else:
            model = SetGNN(data.num_features, num_targets, args)
    elif args.method in ['HGNN','HCHA']:
        model = HCHA(data.num_features, num_targets, args)
    elif args.method == 'HNHN':
        model = HNHN(data.num_features, num_targets, args)
    elif args.method in ['UniGIN']:
        model = UniGNN(data.num_features, num_targets, args)
    elif args.method == 'UniGCNII':
        model = UniGCNII(data.num_features, num_targets, args)
    elif args.method == 'HyperGCN':
        model = HyperGCN(data.num_features, num_targets, args)
    elif args.method == 'LEGCN':
        model = LEGCN(data.num_features, num_targets, args)
    elif args.method == 'HJRL':
        model = HJRL(data.num_features, num_targets, args)
    elif args.method == 'HyperND':
        model = HyperND(data.num_features, num_targets, args)
    elif args.method == 'EDHNN':
        model = EquivSetGNN(data.num_features, num_targets, args)
    elif args.method == 'SheafHyperGNN':
        model = SheafHyperGNN(data.num_features,num_targets,args)
    elif args.method == 'EHNN':
        model = EHNN(data.num_features,num_targets,args,data.ehnn_cache)
    elif args.method == 'TMPHN':
        model = TMPHN(data.num_features,num_targets,data.x,data.neig_dict,args)
    elif args.method == 'PhenomNNS':
        model = PhenomNNS(data.num_features,num_targets,args)
    elif args.method == 'PhenomNN':
        model = PhenomNN(data.num_features,num_targets,args)
    elif args.method == 'DPHGNN':
        model = DPHGNN(data.num_features,num_targets,args)
    elif args.method == 'PlainUnigencoder':
        model = PlainUnigencoder(data.num_features, num_targets, args)
    elif args.method == 'TFHNN':
        model = TFHNN(data.num_features,num_targets,args)
    elif args.method == 'MLP':
        model = PlainMLP(data.num_features,num_targets,args)
    elif args.method == 'HyperGT':
        model = HyperGT(data.num_features,num_targets,args)
    elif args.method == 'CEGCN':
        model = CEGCN(data.num_features,num_targets,args)
    elif args.method == 'CEGAT':
        model = CEGAT(data.num_features,num_targets,args)
    else:
        raise ValueError('Unimplemented model')

    return model


import torch
import torch.nn as nn
import torch.nn.functional as F
from dhgbench.lib_models.HNN.mlp import MLP
import torch_scatter

class EquivSetConv(nn.Module):
    def __init__(self, in_features, out_features, mlp1_layers=1, mlp2_layers=1,
        mlp3_layers=1, aggr='add', alpha=0.5, dropout=0., normalization='None', input_norm=False):
        super().__init__()

        if mlp1_layers > 0:
            self.W1 = MLP(in_features, out_features, out_features, mlp1_layers,
                dropout=dropout, Normalization=normalization, InputNorm=input_norm)
        else:
            self.W1 = nn.Identity()

        if mlp2_layers > 0:
            self.W2 = MLP(in_features+out_features, out_features, out_features, mlp2_layers,
                dropout=dropout, Normalization=normalization, InputNorm=input_norm)
        else:
            self.W2 = lambda X: X[..., in_features:]

        if mlp3_layers > 0:
            self.W = MLP(out_features, out_features, out_features, mlp3_layers,
                dropout=dropout, Normalization=normalization, InputNorm=input_norm)
        else:
            self.W = nn.Identity()
        self.aggr = aggr
        self.alpha = alpha
        self.dropout = dropout

    def reset_parameters(self):
        if isinstance(self.W1, MLP):
            self.W1.reset_parameters()
        if isinstance(self.W2, MLP):
            self.W2.reset_parameters()
        if isinstance(self.W, MLP):
            self.W.reset_parameters()

    def forward(self, X, vertex, edges, X0):
        N = X.shape[-2]

        Xve = self.W1(X)[..., vertex, :] # [nnz, C]
        Xe = torch_scatter.scatter(Xve, edges, dim=-2, reduce=self.aggr) # [E, C], reduce is 'mean' here as default
        
        Xev = Xe[..., edges, :] # [nnz, C]
        Xev = self.W2(torch.cat([X[..., vertex, :], Xev], -1))
        Xv = torch_scatter.scatter(Xev, vertex, dim=-2, reduce=self.aggr, dim_size=N) # [N, C]

        X = Xv

        X = (1-self.alpha) * X + self.alpha * X0
        X = self.W(X)

        return X,Xe

class JumpLinkConv(nn.Module):
    def __init__(self, in_features, out_features, mlp_layers=2, aggr='add', alpha=0.5):
        super().__init__()
        self.W = MLP(in_features, out_features, out_features, mlp_layers,
            dropout=0., Normalization='None', InputNorm=False)

        self.aggr = aggr
        self.alpha = alpha

    def reset_parameters(self):
        self.W.reset_parameters()

    def forward(self, X, vertex, edges, X0, beta=1.):
        N = X.shape[-2]

        Xve = X[..., vertex, :] # [nnz, C]
        Xe = torch_scatter.scatter(Xve, edges, dim=-2, reduce=self.aggr) # [E, C], reduce is 'mean' here as default
        
        Xev = Xe[..., edges, :] # [nnz, C]
        Xv = torch_scatter.scatter(Xev, vertex, dim=-2, reduce=self.aggr, dim_size=N) # [N, C]

        X = Xv

        Xi = (1-self.alpha) * X + self.alpha * X0
        X = (1-beta) * Xi + beta * self.W(Xi)

        return X,Xe

class EquivSetGNN(nn.Module):
    def __init__(self, num_features, num_targets, args):

        super().__init__()
        act = {'Id': nn.Identity(), 'relu': nn.ReLU(), 'prelu':nn.PReLU()}
        self.act = act[args.activation]
        self.dropout = args.dropout # 0.2 is chosen for GCNII
        self.norm = args.normalization

        self.in_channels = num_features
        self.hidden_channels = args.MLP_hidden
        self.output_channels = num_targets

        self.mlp1_layers = args.MLP_num_layers
        self.mlp2_layers = args.MLP_num_layers if args.MLP2_num_layers < 0 else args.MLP2_num_layers
        self.mlp3_layers = args.MLP_num_layers if args.MLP3_num_layers < 0 else args.MLP3_num_layers
        self.num_layers = args.All_num_layers
        self.edconv_type = args.edconv_type

        self.lin_in = torch.nn.Linear(num_features, args.MLP_hidden)
        if args.edconv_type == 'EquivSet':
            self.conv = EquivSetConv(args.MLP_hidden, args.MLP_hidden, mlp1_layers=self.mlp1_layers, mlp2_layers=self.mlp2_layers,
                mlp3_layers=self.mlp3_layers, alpha=args.alpha, aggr=args.aggregate,
                dropout=self.dropout, normalization=self.norm, input_norm=args.AllSet_input_norm)
        elif args.edconv_type == 'JumpLink':
            self.conv = JumpLinkConv(args.MLP_hidden, args.MLP_hidden, mlp_layers=self.mlp1_layers, alpha=args.alpha, aggr=args.aggregate)
        else:
            raise ValueError(f'Unsupported EDConv type: {args.edconv_type}')

        self.decoder = MLP(in_channels=args.MLP_hidden,
            hidden_channels=args.decoder_hidden,
            out_channels=self.output_channels,
            num_layers=args.decoder_num_layer,
            dropout=self.dropout,
            Normalization=self.norm,
            InputNorm=False)

    def reset_parameters(self):
        self.lin_in.reset_parameters()
        self.conv.reset_parameters()
        self.decoder.reset_parameters()

    def forward(self, data):
        x = data.x
        V, E = data.hyperedge_index[0], data.hyperedge_index[1]
        x = F.relu(self.lin_in(x))
        x = F.dropout(x, p=self.dropout, training=self.training)
        x0 = x
        for i in range(self.num_layers):
            x,e = self.conv(x, V, E, x0)
            x = self.act(x)
            x = F.dropout(x, p=self.dropout, training=self.training)
        x = self.decoder(x)
        return x,e



_semi_methods_=['HGNN','HyperGCN','HCHA','HNHN','AllSetformer','AllDeepSets','UniGIN',
          'LEGCN','HyperND','UniGCNII','EDHNN','PlainUnigencoder','HJRL','SheafHyperGNN','EHNN','TMPHN','PhenomNNS','PhenomNN','HyperGT','DPHGNN','TFHNN','CEGCN','CEGAT','MLP'] 